# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import uuid
from datetime import datetime

import mock

from openstackclient.tests.unit import utils

from otcextensions.tests.unit.osclient import test_base

from otcextensions.sdk.nat.v2 import gateway
from otcextensions.sdk.nat.v2 import snat
from otcextensions.sdk.nat.v2 import dnat


def gen_data(data, columns):
    """Fill expected data tuple based on columns list
    """
    return tuple(getattr(data, attr, '') for attr in columns)


def gen_data_dict(data, columns):
    """Fill expected data tuple based on columns list
    """
    return tuple(data.get(attr, '') for attr in columns)


class TestNat(utils.TestCommand):
    def setUp(self):
        super(TestNat, self).setUp()

        self.app.client_manager.nat = mock.Mock()

        self.client = self.app.client_manager.nat


class FakeNatGateway(test_base.Fake):
    """Fake one or more Nat Gateways."""
    @classmethod
    def generate(cls):
        """Create a fake NAT Gateway.

        :return:
            A FakeResource object, with id, name and so on
        """
        # Set default attributes.
        object_info = {
            "id": "id-" + uuid.uuid4().hex,
            "name": "name-" + uuid.uuid4().hex,
            "router_id": "router-" + uuid.uuid4().hex,
            "status": "PENDING_CREATE",
            "description": "my nat gateway",
            "admin_state_up": 'true',
            "tenant_id": "tenant-id-" + uuid.uuid4().hex,
            "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "spec": "1",
            "internal_network_id": "net-id-" + uuid.uuid4().hex
        }

        return gateway.Gateway(**object_info)


class FakeSnatRule(test_base.Fake):
    """Fake one or more SNAT Rule."""
    @classmethod
    def generate(cls):
        """Create a fake SNAT Rule.

        :return:
            A FakeResource object, with id, status and so on
        """
        # Set default attributes.
        object_info = {
            "id": "id-" + uuid.uuid4().hex,
            "floating_ip_id": "eip-id-" + uuid.uuid4().hex,
            "status": "PENDING_CREATE",
            "nat_gateway_id": "gw-id-" + uuid.uuid4().hex,
            "admin_state_up": True,
            "network_id": "net-id-" + uuid.uuid4().hex,
            "cidr": uuid.uuid4().hex,
            "source_type": 0,
            "tenant_id": "tenant-id-" + uuid.uuid4().hex,
            "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "floating_ip_address": uuid.uuid4().hex
        }

        return snat.Snat.existing(**object_info)


class FakeDnatRule(test_base.Fake):
    """Fake one or more DNAT Rule"""
    @classmethod
    def generate(cls):
        """Create a fake DNAT Rule.

        :return:
            A FakeResource object, with id, status and so on
        """
        # Set default attributes.
        object_info = {
            "id": "id-" + uuid.uuid4().hex,
            "floating_ip_id": "eip-id-" + uuid.uuid4().hex,
            "status": "ACTIVE",
            "nat_gateway_id": "gw-id-" + uuid.uuid4().hex,
            "admin_state_up": True,
            "private_ip": uuid.uuid4().hex,
            "internal_service_port": 0,
            "protocol": "any",
            "tenant_id": "abc",
            "port_id": "",
            "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "floating_ip_address": uuid.uuid4().hex,
            "external_service_port": 0
        }

        obj = dnat.Dnat.existing(**object_info)
        return obj



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
"""VPC Endpoint Service v1 action implementations"""
import logging

from osc_lib import exceptions
from osc_lib import utils
from osc_lib.cli import parseractions
from osc_lib.command import command

from otcextensions.common import cli_utils
from otcextensions.common import sdk_utils
from otcextensions.i18n import _

LOG = logging.getLogger(__name__)

SERVER_TYPE_CHOICES = ['vm', 'vip', 'lb']
SERVICE_TYPE_CHOICES = ['gateway', 'interface']


_formatters = {
    'ports': cli_utils.YamlFormat,
    'tags': cli_utils.YamlFormat,
}


def _get_columns(item):
    column_map = {}
    hidden = [
        'location',
    ]
    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden
    )


def translate_response(func):
    def new(self, *args, **kwargs):
        obj = func(self, *args, **kwargs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)

    new.__name__ = func.__name__
    new.__doc__ = func.__doc__
    return new


class ListServices(command.Lister):

    _description = _('List VPC Endpoint Services.')
    columns = (
        'Id',
        'Service Name',
        'Service Type',
        'Server Type',
        'Connection Count',
        'Status',
    )

    def get_parser(self, prog_name):
        parser = super(ListServices, self).get_parser(prog_name)

        parser.add_argument(
            '--id',
            metavar='<id>',
            help=_('ID of the VPC Endpoint Service.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Name of the VPC Endpoint Service.'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Status of the VPC endpoint service.'),
        )
        parser.add_argument(
            '--sort-key',
            metavar='{created_at, updated_at}',
            type=lambda s: s.lower(),
            choices=['created_at', 'updated_at'],
            help=_('Sorting field of the VPC endpoint service list.'),
        )
        parser.add_argument(
            '--sort-dir',
            metavar='{asc, desc}',
            type=lambda s: s.lower(),
            choices=['asc', 'desc'],
            help=_('Sorting order of the VPC endpoint service list.'),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Limit number of VPC endpoint services displayed.'),
        )
        parser.add_argument(
            '--offset',
            metavar='<offset>',
            type=int,
            help=_('Service records after this Offset will be queried.'),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.vpcep
        args_list = [
            'id',
            'name',
            'limit',
            'offset',
            'sort_key',
            'sort_dir',
            'status',
        ]
        attrs = {}
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val

        data = client.services(**attrs)

        return (
            self.columns,
            (utils.get_item_properties(s, self.columns) for s in data),
        )


class ShowService(command.ShowOne):
    _description = _('Show VPC Endpoint Service Details.')

    def get_parser(self, prog_name):
        parser = super(ShowService, self).get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Name or ID of the VPC Endpoint Service.'),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpcep
        return client.find_service(parsed_args.service)


class CreateService(command.ShowOne):
    _description = _('Create new VPC Endpoint Service.')

    def get_parser(self, prog_name):
        parser = super(CreateService, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Specifies name of the Endpoint Service.'),
        )
        parser.add_argument(
            '--port-id',
            metavar='<port_id>',
            required=True,
            help=_(
                'Specify the ID for identifying the backend '
                'resource of the VPC endpoint service.'
            ),
        )
        parser.add_argument(
            '--pool-id',
            metavar='<pool_id>',
            help=_(
                'Specify the ID of the cluster associated with '
                'the target VPCEP resource.'
            ),
        )
        parser.add_argument(
            '--router-id',
            metavar='<router_id>',
            dest='vpc_id',
            required=True,
            help=_(
                'ID of the router (VPC) to which the backend resource '
                'of the VPC endpoint service belongs.'
            ),
        )
        parser.add_argument(
            '--server-type',
            metavar='{LB, VM, VIP, BMS}',
            type=lambda s: s.upper(),
            choices=['LB', 'VM', 'VIP', 'BMS'],
            required=True,
            help=_('Specifies the resource type.'),
        )
        parser.add_argument(
            '--service-type',
            metavar='{gateway, interface}',
            type=lambda s: s.lower(),
            choices=['gateway', 'interface'],
            help=_('Specifies the type of the VPC endpoint service.'),
        )
        parser.add_argument(
            '--ports',
            metavar='client_port=<client-port>,'
            'server_port=<server-port>,'
            'protocol=<protocol>',
            action=parseractions.MultiKeyValueAction,
            required=True,
            required_keys=['client_port', 'server_port', 'protocol'],
            help=_(
                'Example: \n'
                '--ports client_port=8081,server_port=22,protocol=TCP\n'
                'Repeat option to provide multiple ports.'
            ),
        )
        parser.add_argument(
            '--tcp-proxy',
            metavar='{close, toa_open, proxy_open, open, proxy_vni}',
            type=lambda s: s.lower(),
            choices=['close', 'toa_open', 'proxy_open', 'open', 'proxy_vni'],
            help=_(
                'Whether the client IP address and port number or marker_id '
                'information is transmitted to the server.'
            ),
        )
        parser.add_argument(
            '--tags',
            metavar='key=<tag-key>,value=<tag-value>',
            action=parseractions.MultiKeyValueAction,
            dest='tags',
            required_keys=['key', 'value'],
            help=_(
                'Example: \n'
                '--tags key=test-key,value=test-value\n'
                'Repeat option to provide multiple tags.'
            ),
        )
        parser.add_argument(
            '--disable-approval',
            action='store_true',
            help=_('Specifies whether connection approval is required.'),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpcep

        attrs = {'service_name': parsed_args.name}
        args_list = [
            'port_id',
            'pool_id',
            'vpc_id',
            'service_type',
            'server_type',
            'tags',
            'tcp_proxy',
        ]
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val
        ports = []
        for port in parsed_args.ports:
            ports.append(
                {
                    'client_port': int(port['client_port']),
                    'server_port': int(port['server_port']),
                    'protocol': port['protocol'],
                }
            )
        attrs['ports'] = ports

        if parsed_args.disable_approval:
            attrs['approval_enabled'] = False
        return client.create_service(**attrs)


class UpdateService(command.ShowOne):
    _description = _('Update a Endpoint Service.')

    def get_parser(self, prog_name):
        parser = super(UpdateService, self).get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Name or ID of the Vpc endpoint service.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            dest='service_name',
            help=_('Name of the VPC endpoint service.'),
        )
        parser.add_argument(
            '--ports',
            metavar='client_port=<client-port>,'
            'server_port=<server-port>,'
            'protocol=<protocol>',
            action=parseractions.MultiKeyValueAction,
            required_keys=['client_port', 'server_port', 'protocol'],
            help=_(
                'Example: \n'
                '--ports client_port=8081,server_port=22,protocol=TCP\n'
                'Repeat option to provide multiple ports.'
            ),
        )
        parser.add_argument(
            '--port-id',
            metavar='<port_id>',
            help=_(
                'Specify the ID for identifying the backend resource of '
                'the VPC endpoint service.'
            ),
        )
        parser.add_argument(
            '--tcp-proxy',
            metavar='{close, toa_open, proxy_open, open, proxy_vni}',
            type=lambda s: s.lower(),
            choices=['close', 'toa_open', 'proxy_open', 'open', 'proxy_vni'],
            help=_(
                'Whether the client IP address and port number or marker_id '
                'information is transmitted to the server.'
            ),
        )
        approval_group = parser.add_mutually_exclusive_group()
        approval_group.add_argument(
            '--enable-approval',
            action='store_true',
            help=_('Connection approval is required.'),
        )
        approval_group.add_argument(
            '--disable-approval',
            action='store_true',
            help=_('Connection approval is not required.'),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpcep
        args_list = [
            'service_name',
            'port_id',
            'tcp_proxy',
        ]
        attrs = {}
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val
        if parsed_args.ports:
            ports = []
            for port in parsed_args.ports:
                ports.append(
                    {
                        'client_port': int(port['client_port']),
                        'server_port': int(port['server_port']),
                        'protocol': port['protocol'],
                    }
                )
            attrs['ports'] = ports
        if parsed_args.enable_approval:
            attrs['approval_enabled'] = True
        if parsed_args.disable_approval:
            attrs['approval_enabled'] = False
        service = client.find_service(parsed_args.service)
        return client.update_service(service, **attrs)


class DeleteService(command.Command):

    _description = _('Deletes VPC Endpoint Service.')

    def get_parser(self, prog_name):
        parser = super(DeleteService, self).get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            nargs='+',
            help=_('Vpc Endpoint Services(s) to delete (Name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.vpcep
        result = 0
        for service in parsed_args.service:
            try:
                obj = client.find_service(service)
                client.delete_service(obj.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        'Failed to delete Vpc Endpoint Service with '
                        'name or ID "%(service)s": %(e)s'
                    ),
                    {'service': service, 'e': e},
                )
        if result > 0:
            total = len(parsed_args.service)
            msg = _(
                '%(result)s of %(total)s Vpc Endpoint Services(s) failed '
                'to delete.'
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import warnings

from openstack import exceptions
from openstack import service_description

from otcextensions.sdk.tms.v1 import _proxy


class TmsService(service_description.ServiceDescription):
    """The TMS service."""

    supported_versions = {
        '1': _proxy.Proxy
    }

    def _make_proxy(self, conn):
        """Create a Proxy for the tms service.

        :param conn:
          The `openstack.connection.Connection` we're working with.
        """
        config = conn.config
        ep = ''
        region = 'eu-de'
        version_string = config.get_api_version('tms') or '1'
        endpoint_override = config.get_endpoint(self.service_type)

        endpoints = list(conn.identity.endpoints())
        extracted_eps = []
        # Iterate through the endpoints and extract only tms
        for endpoint in endpoints:
            url = endpoint.get("url", "")
            if 'tms' in url:
                extracted_eps.append(url)

        if config.region_name:
            region = config.region_name
        if extracted_eps:
            for url in extracted_eps:
                if region in url:
                    ep = url
        epo = '%(base)s/' % {'base': ep}

        if epo and not endpoint_override:
            endpoint_override = epo

        proxy_obj = None
        if endpoint_override and self.supported_versions:
            proxy_class = self.supported_versions.get(version_string[0])
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
                proxy_obj.endpoint_override = endpoint_override
                proxy_obj.additional_headers = {
                    'Content-Type': 'application/json'}
            else:
                warnings.warn(
                    "The configured version, {version} for service"
                    " {service_type} is not known or supported by"
                    " openstacksdk. The resulting Proxy object will only"
                    " have direct passthrough REST capabilities.".format(
                        version=version_string,
                        service_type=self.service_type),
                    category=exceptions.UnsupportedServiceVersion)
        elif endpoint_override and self.supported_versions:
            temp_adapter = config.get_session_client(
                self.service_type
            )
            api_version = temp_adapter.get_endpoint_data().api_version
            proxy_class = self.supported_versions.get(str(api_version[0]))
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            else:
                warnings.warn(
                    "Service {service_type} has an endpoint override set"
                    " but the version discovered at that endpoint, {version}"
                    " is not supported by openstacksdk. The resulting Proxy"
                    " object will only have direct passthrough REST"
                    " capabilities.".format(
                        version=api_version,
                        service_type=self.service_type),
                    category=exceptions.UnsupportedServiceVersion)

        if proxy_obj:
            if getattr(proxy_obj, 'skip_discovery', False):
                # Some services, like swift, don't have discovery. While
                # keystoneauth will behave correctly and handle such
                # scenarios, it's not super efficient as it involves trying
                # and falling back a few times.
                return proxy_obj

            data = proxy_obj.get_endpoint_data()
            # If we've gotten here with a proxy object it means we have
            # an endpoint_override in place. If the catalog_url and
            # service_url don't match, which can happen if there is a
            # None plugin and auth.endpoint like with standalone ironic,
            # we need to be explicit that this service has an endpoint_override
            # so that subsequent discovery calls don't get made incorrectly.
            if data.catalog_url != data.service_url:
                ep_key = '{service_type}_endpoint_override'.format(
                    service_type=self.service_type)
                config.config[ep_key] = data.service_url
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            return proxy_obj

        # Make an adapter to let discovery take over
        version_kwargs = {}
        if version_string:
            version_kwargs['version'] = version_string
        elif self.supported_versions:
            supported_versions = sorted([
                int(f) for f in self.supported_versions])
            version_kwargs['min_version'] = str(supported_versions[0])
            version_kwargs['max_version'] = '{version}.latest'.format(
                version=str(supported_versions[-1]))

        temp_adapter = config.get_session_client(
            self.service_type,
            allow_version_hack=True,
            **version_kwargs
        )
        found_version = temp_adapter.get_api_major_version()
        if found_version is None:
            if version_kwargs:
                raise exceptions.NotSupported(
                    "The {service_type} service for {cloud}:{region_name}"
                    " exists but does not have any supported versions.".format(
                        service_type=self.service_type,
                        cloud=conn.name,
                        region_name=conn.config.region_name))
            else:
                raise exceptions.NotSupported(
                    "The {service_type} service for {cloud}:{region_name}"
                    " exists but no version was discoverable.".format(
                        service_type=self.service_type,
                        cloud=conn.name,
                        region_name=conn.config.region_name))
        proxy_class = self.supported_versions.get(str(found_version[0]))
        if not proxy_class:
            # Maybe openstacksdk is being used for the passthrough
            # REST API proxy layer for an unknown service in the
            # service catalog that also doesn't have any useful
            # version discovery?
            warnings.warn(
                "Service {service_type} has no discoverable version."
                " The resulting Proxy object will only have direct"
                " passthrough REST capabilities.".format(
                    service_type=self.service_type),
                category=exceptions.UnsupportedServiceVersion)
            return temp_adapter
        proxy_class = self.supported_versions.get(str(found_version[0]))
        if proxy_class:
            version_kwargs['constructor'] = proxy_class
        return config.get_session_client(
            self.service_type,
            allow_version_hack=True,
            **version_kwargs
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings
from openstack import exceptions
from openstack import warnings as os_warnings


class ServiceProxy:
    def __init__(self, instance, service_type, supported_versions):
        self.instance = instance
        self.service_type = service_type
        self.supported_versions = supported_versions

    def create_proxy(self, base_service="aomv2", target_service="apig"):
        config = self.instance.config

        # Retrieve API version and endpoint configuration
        version_string = config.get_api_version(base_service) or '2'
        endpoint_override = config.get_endpoint(self.service_type)

        # Fetch service catalog URL and adjust the endpoint if necessary
        ep = config.get_service_catalog().url_for(
            service_type=base_service,
            region_name=config.region_name
        )
        epo = f"{ep.replace(base_service.split('v')[0], target_service)}"

        if epo and not endpoint_override:
            endpoint_override = epo

        if not version_string and len(self.supported_versions) == 1:
            version_string = list(self.supported_versions)[0]

        # Attempt to create a proxy object with the
        # specified endpoint and version
        proxy_obj = self._get_proxy_object(
            config,
            version_string,
            endpoint_override
        )
        if proxy_obj:
            return self._handle_discovery(proxy_obj, config)

        # If no specific endpoint or version,
        # allow discovery with version constraints
        return self._create_discovery_adapter(config, version_string)

    def _get_proxy_object(self, config, version_string, endpoint_override):
        proxy_class = None
        if endpoint_override and version_string and self.supported_versions:
            proxy_class = self.supported_versions.get(version_string[0])
            if proxy_class:
                return self._construct_proxy(
                    config,
                    proxy_class,
                    endpoint_override
                )
            else:
                warnings.warn(
                    f"The configured version {version_string} "
                    f"for service {self.service_type} "
                    f"is not known or supported. The resulting "
                    f"Proxy object will only have direct "
                    f"passthrough REST capabilities.",
                    category=os_warnings.UnsupportedServiceVersion)
        elif endpoint_override and self.supported_versions:
            temp_adapter = config.get_session_client(self.service_type)
            api_version = temp_adapter.get_endpoint_data().api_version
            proxy_class = self.supported_versions.get(str(api_version[0]))
            if proxy_class:
                return self._construct_proxy(
                    config,
                    proxy_class,
                    endpoint_override
                )
            else:
                warnings.warn(
                    f"Service {self.service_type} has an "
                    f"endpoint override set, "
                    f"but the version discovered ({api_version}) is "
                    f"not supported. The resulting Proxy object will "
                    f"only have direct passthrough REST capabilities.",
                    category=os_warnings.UnsupportedServiceVersion)

    def _construct_proxy(self, config, proxy_class, endpoint_override):
        proxy_obj = config.get_session_client(
            self.service_type,
            constructor=proxy_class,
        )
        proxy_obj.endpoint_override = endpoint_override
        proxy_obj.additional_headers = {'Content-Type': 'application/json'}
        return proxy_obj

    def _handle_discovery(self, proxy_obj, config):
        if getattr(proxy_obj, 'skip_discovery', False):
            return proxy_obj

        data = proxy_obj.get_endpoint_data()
        if data.catalog_url != data.service_url:
            ep_key = f'{self.service_type}_endpoint_override'
            config.config[ep_key] = data.service_url
            proxy_obj = config.get_session_client(self.service_type)
        return proxy_obj

    def _create_discovery_adapter(self, config, version_string):
        version_kwargs = {}
        if version_string:
            version_kwargs['version'] = version_string
        elif self.supported_versions:
            sv = sorted(int(f) for f in self.supported_versions)
            version_kwargs['min_version'] = str(sv[0])
            version_kwargs['max_version'] = f"{sv[-1]}.latest"

        temp_adapter = config.get_session_client(
            self.service_type,
            allow_version_hack=True,
            **version_kwargs
        )
        found_version = temp_adapter.get_api_major_version()
        if found_version is None:
            raise exceptions.NotSupported(
                f"The {self.service_type} service for "
                f"{self.instance.name}:{config.region_name} "
                f"exists but does not have any supported versions."
            )
        proxy_class = self.supported_versions.get(str(found_version[0]))
        if proxy_class:
            version_kwargs['constructor'] = proxy_class
        else:
            warnings.warn(
                f"Service {self.service_type} has no discoverable version. "
                f"The resulting Proxy object will only have direct"
                f" passthrough REST capabilities.",
                category=os_warnings.UnsupportedServiceVersion)
        return config.get_session_client(
            self.service_type,
            allow_version_hack=True, **version_kwargs
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource
from openstack import exceptions


class Alias(resource.Resource):
    base_path = '/fgs/functions/%(function_urn)s/aliases'
    resources_key = ''

    # Capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True
    allow_commit = True

    # Properties
    function_urn = resource.URI('function_urn', type=str)
    name = resource.Body('name', type=str, alternate_id=True)
    version = resource.Body('version', type=str)
    description = resource.Body('description', type=str)
    additional_version_weights = resource.Body(
        'additional_version_weights', type=dict
    )
    additional_version_strategy = resource.Body(
        'additional_version_strategy', type=dict
    )

    # Attributes
    last_modified = resource.Body('last_modified', type=str)
    alias_urn = resource.Body('alias_urn', type=str)

    def _update_alias(self, session, function, alias, **attrs):
        """Update Function Alias
        """
        urn = function.func_urn.rpartition(':')[0]
        url = self.base_path % {'function_urn': urn} + f'/{alias.id}'
        response = session.put(url, json=attrs)
        exceptions.raise_from_response(response)
        self._translate_response(response)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource
from openstack import utils


class RecordSpec(resource.Resource):
    #: Data to be uploaded.
    data = resource.Body('data')
    #: Hash value of the data to be written to the partition.
    #:  The hash value overwrites the hash value of partition_key.
    explicit_hash_key = resource.Body('explicit_hash_key')
    #: Partition ID of the stream.
    partition_id = resource.Body('partition_id')
    #: Partition to which data is written.
    partition_key = resource.Body('partition_key')


class Data(resource.Resource):
    base_path = '/records'

    resources_key = 'records'

    _query_mapping = resource.QueryParameters(
        # 'stream-name', 'partition-id', 'cursor-type',
        # 'starting-sequence-number', 'timestamp', 'stream_id',
        'partition-cursor', 'max_fetch_bytes',
    )

    allow_create = True
    allow_fetch = True
    allow_list = True

    # Properties
    #: Partition key set when data is being uploaded.
    partition_key = resource.Body('partition_key')
    #: Sequence number of the data record.
    sequence_number = resource.Body('sequence_number')
    #: Downloaded data. The downloaded data is the
    #:  serialized binary data (Base64-encoded character string).
    data = resource.Body('data')
    #: Data cursor
    #: \nValue: 1 to 512 characters
    #: \nNote:
    #: \nThe validity period of a data cursor is 5 minutes.
    #: \nMinimum: 1
    #: \nMaximum: 512
    partition_cursor = resource.Body('partition_cursor')
    #: Timestamp when the record is written to DIS.
    timestamp = resource.Body('timestamp', type=int)
    #: Timestamp type.
    timestamp_type = resource.Body('timestamp_type')
    #: List of records to be uploaded.
    records = resource.Body('records', type=list, list_type=RecordSpec)
    #: Unique ID of the stream. If no stream is found based on stream_name
    #:  and stream_id is not empty, stream_id is used to search for the stream.
    stream_id = resource.Body('stream_id')
    #: Name of the stream.
    stream_name = resource.Body('stream_name')

    def get_data_cursor(self, session, **params):
        uri = utils.urljoin('cursors')
        res = session.get(uri, params=params)
        self._translate_response(res)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
"""VPC Peering v2 action implementations"""
import logging

from osc_lib import utils
from osc_lib import exceptions
from osc_lib.command import command

from otcextensions.i18n import _

LOG = logging.getLogger(__name__)

STATUS_CHOICES = [
    'PENDING_ACCEPTANCE',
    'REJECTED',
    'EXPIRED',
    'DELETED',
    'ACTIVE'
]


def _update_vpc_info(vpcinfo):
    if 'vpc_id' in vpcinfo:
        vpcinfo['router_id'] = vpcinfo.pop('vpc_id')
    if 'tenant_id' in vpcinfo:
        vpcinfo['project_id'] = vpcinfo.pop('tenant_id')
    return vpcinfo


def set_attributes_for_print(peerings):
    for peering in peerings:
        setattr(peering, 'local_router_id',
                peering.local_vpc_info['vpc_id'])
        setattr(peering, 'peer_router_id',
                peering.peer_vpc_info['vpc_id'])
        setattr(peering, 'peer_project_id',
                peering.peer_vpc_info['tenant_id'])
        yield peering


def translate_response(func):
    def new(self, *args, **kwargs):
        response = func(self, *args, **kwargs)
        setattr(response,
                'local_vpc_info',
                _update_vpc_info(response.local_vpc_info))
        setattr(response,
                'peer_vpc_info',
                _update_vpc_info(response.peer_vpc_info))
        columns = (
            'id',
            'name',
            'local_vpc_info',
            'peer_vpc_info',
            'description',
            'created_at',
            'updated_at',
            'status'
        )
        data = utils.get_item_properties(response, columns)
        return (columns, data)
    new.__name__ = func.__name__
    new.__doc__ = func.__doc__
    return new


class ListVpcPeerings(command.Lister):

    _description = _("List Vpc Peerings.")
    columns = (
        'Id',
        'Name',
        'Status',
        'Local Router Id',
        'Peer Router Id',
        'Peer Project Id')

    def get_parser(self, prog_name):
        parser = super(ListVpcPeerings, self).get_parser(prog_name)

        parser.add_argument(
            '--id',
            metavar='<id>',
            help=_("Specifies the ID of the VPC peering connection."),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_("Limit to fetch number of records."),
        )
        parser.add_argument(
            '--marker',
            metavar='<marker>',
            help=_("Specifies the start resource ID of pagination query."),
        )
        parser.add_argument(
            '--project-id',
            metavar='<project_id>',
            help=_("Specifies the project ID."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_("Specifies the Name of the VPC peering connection."),
        )
        parser.add_argument(
            '--router-id',
            metavar='<router_id>',
            help=_("Specifies the router/vpc ID of requester/accepter "
                   "of vpc peering."),
        )
        parser.add_argument(
            '--status',
            metavar='{' + ','.join(STATUS_CHOICES) + '}',
            type=lambda s: s.upper(),
            choices=STATUS_CHOICES,
            help=_("Specifies the status of the VPC peering connection."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        args_list = [
            'id',
            'name',
            'limit',
            'marker',
            'project_id',
            'router_id',
            'status']
        attrs = {}
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val

        data = client.peerings(**attrs)
        if data:
            data = set_attributes_for_print(data)

        return (self.columns, (utils.get_item_properties(s, self.columns)
                               for s in data))


class ShowVpcPeering(command.ShowOne):
    _description = _("Show VPC peering connection details.")

    def get_parser(self, prog_name):
        parser = super(ShowVpcPeering, self).get_parser(prog_name)
        parser.add_argument(
            'peering',
            metavar='<peering>',
            help=_("Specifies the Name or ID of the VPC peering connection."),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        return client.find_peering(parsed_args.peering)


class SetVpcPeering(command.ShowOne):
    _description = _("Accept VPC peering connection Request.")

    def get_parser(self, prog_name):
        parser = super(SetVpcPeering, self).get_parser(prog_name)
        parser.add_argument(
            'peering',
            metavar='<peering>',
            help=_("Specifies the Name or ID of the VPC peering connection."),
        )
        manage_request_group = parser.add_mutually_exclusive_group()
        manage_request_group.add_argument(
            '--accept',
            action='store_true',
            help=('Accept VPC peering connection request.')
        )
        manage_request_group.add_argument(
            '--reject',
            action='store_true',
            help=('Reject VPC peering connection request.')
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        set_args = ('accept', 'reject')
        request_status = [request for request in set_args if
                          getattr(parsed_args, request)]
        peering = client.find_peering(parsed_args.peering)
        return client.set_peering(peering.id, request_status[0])


class CreateVpcPeering(command.ShowOne):
    _description = _("Create new Vpc Peering.")

    def get_parser(self, prog_name):
        parser = super(CreateVpcPeering, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_("Specifies the name of the VPC peering connection."),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Specifies the description of the VPC peering connection."),
        )
        parser.add_argument(
            '--local-router-id',
            metavar='<local_router_id>',
            required=True,
            help=_("Specifies information about the local router_id "
                   "involved in a VPC peering connection."),
        )
        parser.add_argument(
            '--peer-router-id',
            metavar='<peer_router_id>',
            required=True,
            help=_("Specifies information about the peer router_id "
                   "involved in a VPC peering connection."),
        )
        parser.add_argument(
            '--peer-project-id',
            metavar='<peer_project_id>',
            help=_("Specifies the ID of the project to which a "
                   "peer router belongs. It is required if the peer "
                   "router is from different project."),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        attrs = {
            'name': parsed_args.name,
            'request_vpc_info': {
                'vpc_id': parsed_args.local_router_id
            },
            'accept_vpc_info': {
                'vpc_id': parsed_args.peer_router_id
            }
        }
        val = getattr(parsed_args, 'peer_project_id')
        if val:
            attrs['accept_vpc_info']['tenant_id'] = val
            attrs['request_vpc_info']['tenant_id'] = client.get_project_id()

        val = getattr(parsed_args, 'description')
        if val:
            attrs['description'] = val

        return client.create_peering(**attrs)


class UpdateVpcPeering(command.ShowOne):
    _description = _("Update a VPC peering connection.")

    def get_parser(self, prog_name):
        parser = super(UpdateVpcPeering, self).get_parser(prog_name)
        parser.add_argument(
            'peering',
            metavar='<peering>',
            help=_("Specifies the Name or ID of the VPC peering connection."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_("Specifies the new name of the VPC peering connection."),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Specifies the new descrition of the VPC peering "
                   "connection."),
        )
        return parser

    @translate_response
    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        args_list = [
            'name', 'description'
        ]
        attrs = {}
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val
        peering = client.find_peering(parsed_args.peering)

        return client.update_peering(peering.id, **attrs)


class DeleteVpcPeering(command.Command):

    _description = _("Deletes VPC Peering.")

    def get_parser(self, prog_name):
        parser = super(DeleteVpcPeering, self).get_parser(prog_name)
        parser.add_argument(
            'peering',
            metavar='<peering>',
            nargs='+',
            help=_("VPC Peering(s) to delete (Name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.vpc
        result = 0
        for peering in parsed_args.peering:
            try:
                obj = client.find_peering(peering)
                client.delete_peering(obj.id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete VPC peering connection with "
                          "name or ID '%(peering)s': %(e)s"),
                          {'peering': peering, 'e': e})
        if result > 0:
            total = len(parsed_args.peering)
            msg = (_("%(result)s of %(total)s VPC peering(s) failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import uuid
import mock
from datetime import datetime

from openstackclient.tests.unit import utils
from otcextensions.tests.unit.osclient import test_base

from otcextensions.sdk.vpc.v1 import peering
from otcextensions.sdk.vpc.v1 import route


def gen_data(data, columns):
    """Fill expected data tuple based on columns list
    """
    return tuple(getattr(data, attr, '') for attr in columns)


def gen_data_dict(data, columns):
    """Fill expected data tuple based on columns list
    """
    return tuple(data.get(attr, '') for attr in columns)


class TestVpc(utils.TestCommand):
    def setUp(self):
        super(TestVpc, self).setUp()

        self.app.client_manager.vpc = mock.Mock()

        self.client = self.app.client_manager.vpc


class FakeVpcPeering(test_base.Fake):
    """Fake one or more VPC peering connections."""
    @classmethod
    def generate(cls):
        """Create a fake VPC peering connection.

        :return:
            A FakeResource object, with id, name and so on
        """
        # Set default attributes.
        object_info = {
            "id": "id-" + uuid.uuid4().hex,
            "name": "name-" + uuid.uuid4().hex,
            "request_vpc_info": {
                "vpc_id": uuid.uuid4().hex,
                "tenant_id": uuid.uuid4().hex
            },
            "accept_vpc_info": {
                "vpc_id": uuid.uuid4().hex,
                "tenant_id": uuid.uuid4().hex
            },
            "status": "ACTIVE",
            "description": "my vpc peering",
            "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "updated_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        }

        return peering.Peering(**object_info)


class FakeVpcRoute(test_base.Fake):
    """Fake one or more VPC routes."""
    @classmethod
    def generate(cls):
        """Create a fake VPC route.

        :return:
            A FakeResource object, with id, name and so on
        """
        # Set default attributes.
        object_info = {
            "id": "id-" + uuid.uuid4().hex,
            "type": "peering",
            "nexthop": uuid.uuid4().hex,
            "router_id": uuid.uuid4().hex,
            "project_id": uuid.uuid4().hex,
            "destination": "192.168.200.0/24",
        }

        return route.Route(**object_info)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
from urllib import parse
from urllib.parse import urlsplit

from openstack import exceptions
from urllib3.exceptions import LocationParseError

from otcextensions.common import utils
from otcextensions.common.utils import extract_region_from_url
from otcextensions.sdk import ak_auth
from otcextensions.sdk import sdk_proxy
from otcextensions.sdk.obs.v1 import container as _container
from otcextensions.sdk.obs.v1 import obj as _obj

DEFAULT_OBJECT_SEGMENT_SIZE = 1073741824  # 1GB
DEFAULT_MAX_FILE_SIZE = int((5 * 1024 * 1024 * 1024 + 2) / 2)
EXPIRES_ISO8601_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
SHORT_EXPIRES_ISO8601_FORMAT = '%Y-%m-%d'


def _normalize_obs_keys(obj):
    return {k.lower(): v for k, v in obj.items()}


class Proxy(sdk_proxy.Proxy):
    skip_discovery = True

    CONTAINER_ENDPOINT = \
        'https://%(container)s.obs.%(region_name)s.otc.t-systems.com'

    def _extract_name(self, url, service_type=None, project_id=None):
        url_path = parse.urlparse(url).path.strip()
        # Remove / from the beginning to keep the list indexes of interesting
        # things consistent
        if url_path.startswith('/'):
            url_path = url_path[1:]

        # Split url into parts and exclude potential project_id in some urls
        url_parts = url_path.split('/')

        # Strip out anything that's empty or None
        parts = [part for part in url_parts if part]

        # Getting the root of an endpoint is a bucket operation
        if not parts:
            return ['bucket']
        else:
            return ['object']

    def get_container_endpoint(self, container):
        """Override to return mapped endpoint if override and region are set

        """
        split_url = urlsplit(self.get_endpoint())

        return f'{split_url.scheme}://%(container)s.{split_url.netloc}' % \
               {'container': container}

    def _get_req_auth(self, host=None):
        auth = getattr(self, '_ak_auth', None)
        if not auth:
            ak = None
            sk = None
            token = None
            conn = self.session._sdk_connection
            if hasattr(conn, 'get_ak_sk'):
                aksk = conn.get_ak_sk(conn)
                if len(aksk) == 2:
                    (ak, sk) = aksk
                elif len(aksk) == 3:
                    (ak, sk, token) = aksk
            if not (ak and sk):
                self.log.error('Cannot obtain AK/SK from config')
                return None
            region = extract_region_from_url(self.get_endpoint())
            if not host:
                host = self.get_endpoint()
            auth_params = {
                "access_key": ak,
                "secret_access_key": sk,
                "host": host,
                "region": region,
                "service": "s3"
            }
            if token:
                auth_params["token"] = token
            auth = ak_auth.AKRequestsAuth(**auth_params)
            setattr(self, '_ak_auth', auth)
        return auth

    # ======== Containers ========

    def containers(self, **query):
        """Obtain Container objects for this account.

        :param kwargs query: Optional query parameters to be sent to limit
                                 the resources being returned.

        :rtype: A generator of
            :class:`~otcextensions.sdk.obs.v1.container.Container` objects.
        """
        return self._list(_container.Container,
                          requests_auth=self._get_req_auth(),
                          **query)

    def get_container(self, container):
        """Get the detail of a container

        :param container: Container name or an object of class
                   :class:`~otcextensions.sdk.obs.v1.container.Container`
        :returns: Detail of container
        :rtype: :class:`~otcextensions.sdk.obs.v1.container.Container`
        """
        container = self._get_container_name(container=container)
        endpoint = self.get_container_endpoint(container)
        return self._head(
            _container.Container,
            container,
            endpoint_override=endpoint,
            requests_auth=self._get_req_auth(endpoint))

    def create_container(self, **attrs):
        """Create a new container from attributes

        :param container: Name of the container to create.
        :param dict attrs: Keyword arguments which will be used to create
               a :class:`~otcextensions.sdk.obs.v1.container.Container`,
               comprised of the properties on the Container class.

        :returns: The results of container creation
        :rtype: :class:`~otcextensions.sdk.obs.v1.container.Container`
        """
        container = attrs.get('name', None)
        endpoint = self.get_container_endpoint(container)
        request_auth = self._get_req_auth(endpoint)
        return self._create(
            _container.Container,
            endpoint_override=endpoint,
            requests_auth=request_auth,
            # prepend_key=False,
            **attrs)

    def delete_container(self, container, ignore_missing=True):
        """Delete a container

        :param container: The value can be either the name of a container or a
                      :class:`~otcextensions.sdk.obs.v1.container.Container`
                      instance.
        :param bool ignore_missing: When set to ``False``
                    :class:`~openstack.exceptions.ResourceNotFound` will be
                    raised when the container does not exist.
                    When set to ``True``, no exception will be set when
                    attempting to delete a nonexistent server.

        :returns: ``None``
        """
        container = self._get_container_name(container=container)
        endpoint = self.get_container_endpoint(container)
        self._delete(_container.Container, container,
                     endpoint_override=endpoint,
                     requests_auth=self._get_req_auth(endpoint),
                     ignore_missing=ignore_missing)

    def get_container_metadata(self, container):
        """Get metadata for a container

        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.

        :returns: One :class:`~otcextensions.sdk.obs.v1.container.Container`
        :raises: :class:`~openstack.exceptions.ResourceNotFound`
                 when no resource can be found.
        """
        raise NotImplementedError
        return self._head(_container.Container, container)

    def set_container_metadata(self, container, **metadata):
        """Set metadata for a container.

        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param kwargs metadata: Key/value pairs to be set as metadata
                                on the container. Both custom and system
                                metadata can be set. Custom metadata are keys
                                and values defined by the user. System
                                metadata are keys defined by the Object Store
                                and values defined by the user. The system
                                metadata keys are:

                                - `content_type`
                                - `is_content_type_detected`
                                - `versions_location`
                                - `read_ACL`
                                - `write_ACL`
                                - `sync_to`
                                - `sync_key`
        """
        raise NotImplementedError
        res = self._get_resource(_container.Container, container)
        res.set_metadata(self, metadata)
        return res

    def delete_container_metadata(self, container, keys):
        """Delete metadata for a container.

        :param container: The value can be the ID of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param keys: The keys of metadata to be deleted.
        """
        raise NotImplementedError
        res = self._get_resource(_container.Container, container)
        res.delete_metadata(self, keys)
        return res

    # ======== Objects ========

    def objects(self, container, **query):
        """Return a generator that yields the Container's objects.

        :param container: A container object or the name of a container
            that you want to retrieve objects from.
        :type container:
            :class:`~otcextensions.sdk.obs.v1.container.Container`
        :param kwargs query: Optional query parameters to be sent to limit
                               the resources being returned.

        :rtype: A generator of
            :class:`~otcextensions.sdk.obs.v1.obj.Object` objects.
        """
        container = self._get_container_name(container=container)
        endpoint = self.get_container_endpoint(container)
        return self._list(
            _obj.Object,
            endpoint_override=endpoint,
            requests_auth=self._get_req_auth(endpoint), **query)

    def _get_container_name(self, obj=None, container=None):
        if obj is not None:
            obj = self._get_resource(_obj.Object, obj)
            if obj.container is not None:
                return obj.container
        if container is not None:
            container = self._get_resource(_container.Container, container)
            return container.name

        raise ValueError("container must be specified")

    def get_object(self, obj, container=None):
        """Get the data associated with an object

        :param obj: The value can be the name of an object or a
                       :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.

        :returns: The contents of the object.  Use the
                  :func:`~get_object_metadata`
                  method if you want an object resource.
        :raises: :class:`~openstack.exceptions.ResourceNotFound`
                 when no resource can be found.
        """
        container_name = self._get_container_name(
            obj=obj, container=container)
        endpoint = self.get_container_endpoint(container_name)
        return self._get(_obj.Object, obj, container=container_name,
                         endpoint_override=endpoint,
                         requests_auth=self._get_req_auth(endpoint))

    def download_object(self, obj, container=None, **attrs):
        """Download the data contained inside an object.

        :param obj: The value can be the name of an object or a
                       :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.

        :raises: :class:`~openstack.exceptions.ResourceNotFound`
                 when no resource can be found.
        """
        container_name = self._get_container_name(
            obj=obj, container=container)
        endpoint = self.get_container_endpoint(container_name)
        obj = self._get_resource(
            _obj.Object, obj, container=container_name, **attrs)
        return obj.download(
            self,
            endpoint_override=endpoint,
            requests_auth=self._get_req_auth(endpoint),
            filename=attrs.pop('file', '-'))

    def stream_object(self, obj, container=None, chunk_size=1024, **attrs):
        """Stream the data contained inside an object.

        :param obj: The value can be the name of an object or a
                       :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.

        :raises: :class:`~openstack.exceptions.ResourceNotFound`
                 when no resource can be found.
        :returns: An iterator that iterates over chunk_size bytes
        """
        raise NotImplementedError
        # container_name = self._get_container_name(
        #     obj=obj, container=container)
        container_name = self._get_container_name(container=container)
        endpoint = self.get_container_endpoint(container_name)
        obj = self._get_resource(
            _obj.Object, obj, container=container_name,
            endpoint_override=endpoint,
            requests_auth=self._get_req_auth(endpoint),
            **attrs)
        return obj.stream(self, chunk_size=chunk_size)

    def create_object(self, container, name, filename=None, data=None,
                      segment_size=None, md5=None,
                      generate_checksums=None,
                      **headers):
        """Upload a new object from attributes

        :param generate_checksums: Whether to generate checksums on the client
            side that get added to headers for later prevention of double
            uploads of identical data. (optional, defaults to True)
        :param md5: A hexadecimal md5 of the file. (Optional), if it is known
            and can be passed here, it will save repeating the expensive md5
            process. It is assumed to be accurate.
        :param segment_size: Break the uploaded object into segments of this
            many bytes. (Optional) SDK will attempt to discover the maximum
            value for this from the server if it is not specified, or will use
            a reasonable default.
        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param name: Name of the object to create.
        :param data: The content to upload to the object. Mutually exclusive
            with filename.
        :param filename: The path to the local file whose contents will be
            uploaded. Mutually exclusive with data.
        :param dict attrs: Keyword arguments which will be used to create
               a :class:`~otcextensions.sdk.obs.v1.obj.Object`,
               comprised of the properties on the Object class.

        :returns: The results of object creation
        :rtype: :class:`~otcextensions.sdk.obs.v1.container.Container`
        """
        if data is not None and filename:
            raise ValueError(
                "Both filename and data given. Please choose one.")
        if data is not None and not name:
            raise ValueError("name is a required parameter when data is given")
        if data is not None and generate_checksums:
            raise ValueError(
                "checksums cannot be generated with data parameter"
            )

        if not filename and data is None:
            filename = name

        container = self._get_container_name(container=container)
        endpoint = self.get_container_endpoint(container)

        # folder logic
        if data is None and name[-1] == "/":
            return self._create(_obj.Object, container=container,
                                name=name,
                                endpoint_override=endpoint,
                                requests_auth=self._get_req_auth(endpoint),
                                **headers)

        if data is not None:
            self.log.debug(
                "uploading data to %(endpoint)s",
                {'endpoint': endpoint})
            return self._create(
                _obj.Object, container=container,
                name=name, data=data,
                endpoint_override=endpoint,
                requests_auth=self._get_req_auth(endpoint),
                **headers)

        if segment_size:
            segment_size = int(segment_size)
        segment_size = self.get_object_segment_size(segment_size)
        file_size = os.path.getsize(filename)

        if generate_checksums and md5 is None:
            md5 = utils._get_file_hashes(filename)

        if self.is_object_stale(container, name, filename, md5):

            self.log.debug(
                "uploading %(filename)s to %(endpoint)s",
                {'filename': filename, 'endpoint': endpoint})

            if file_size <= segment_size:
                self._upload_object(endpoint, filename, headers, name)
            else:
                self._upload_large_object(
                    endpoint, filename, name, headers,
                    file_size, segment_size)

    # Backwards compat
    upload_object = create_object

    def _upload_object(self, endpoint, filename, headers, name=None):
        if not name:
            name = os.path.basename(filename)
        with open(filename, 'rb') as dt:
            return self._create(
                _obj.Object,
                name=name, data=dt,
                endpoint_override=endpoint,
                requests_auth=self._get_req_auth(endpoint),
                **headers)

    def _upload_large_object(self, endpoint, filename, name, headers,
                             file_size, segment_size):
        """
        If the object is big, we need to break it up into segments that
        are no larger than segment_size, upload each of them individually
        and then upload a manifest object. The segments can be uploaded in
        parallel, so we'll use the async feature of the TaskManager.
        """

        segment_futures = []
        segment_results = []
        retry_results = []
        retry_futures = []
        manifest = []
        object_name = os.path.basename(filename)
        if name:
            object_name = name
        requests_auth = self._get_req_auth(endpoint)
        segments = utils._get_file_segments(
            endpoint, filename, file_size, segment_size)

        upload_id = _obj.Object.initiate_multipart_upload(
            self, endpoint, object_name,
            requests_auth=requests_auth
        )
        url = f'{endpoint}/{object_name}'
        # Schedule the segments for upload
        for name, segment in segments.items():
            part_number = name.rsplit('/', 1)[-1]
            # Async call to put - schedules execution and returns a future
            segment_future = self._connection._pool_executor.submit(
                self.put,
                f'{url}?partNumber={part_number}&uploadId={upload_id}',
                headers=headers, data=segment,
                requests_auth=requests_auth,
                raise_exc=False)
            segment_futures.append(segment_future)
            # dict. Then sort the list of dicts by path.
            manifest.append(dict(
                path='/{name}'.format(name=name),
                size_bytes=segment.length))

        segment_results, retry_results = self._connection._wait_for_futures(
            segment_futures, raise_on_error=False)

        for result in retry_results:
            # Grab the FileSegment for the failed upload so we can retry
            name = self._object_name_from_url(result.url)
            segment = segments[name]
            segment.seek(0)
            # Async call to put - schedules execution and returns a future
            segment_future = self._connection._pool_executor.submit(
                self.put,
                f'{url}?partNumber={name[-1]}&uploadId={upload_id}',
                headers=headers, data=segment,
                requests_auth=requests_auth
            )
            # dict. Then sort the list of dicts by path.
            retry_futures.append(segment_future)

        # If any segments fail the second time, just throw the error
        segment_results, retry_results = self._connection._wait_for_futures(
            retry_futures, raise_on_error=True)

        try:
            return self._finish_large_object_upload(
                url, headers, upload_id)
        except Exception:
            try:
                self.log.debug(
                    "Failed to upload large object for %s. "
                    "Aborting uploads.", upload_id)
                self._abort_multipart_upload(endpoint=url,
                                             upload_id=upload_id)
            except Exception:
                self.log.exception(
                    "Failed to cleanup image objects for %s:",
                    upload_id)
            raise

    def parts(self, endpoint, upload_id, requests_auth):
        return _obj.Object.get_parts(
            self, f'{endpoint}?uploadId={upload_id}', requests_auth)

    def _finish_large_object_upload(self, endpoint, headers, upload_id):
        requests_auth = self._get_req_auth(endpoint)
        parts = self.parts(endpoint, upload_id, requests_auth)
        retries = 3
        while True:
            try:
                return exceptions.raise_from_response(
                    _obj.Object.complete_multipart_upload(
                        self, endpoint, upload_id,
                        parts['Parts'], headers,
                        requests_auth=requests_auth
                    ))
            except Exception:
                retries -= 1
                if retries == 0:
                    raise

    def _object_name_from_url(self, url):
        '''Get container_name/object_name from the full URL called.
        Remove the Swift endpoint from the front of the URL, and remove
        the leaving / that will leave behind.'''
        endpoint = self.get_endpoint()
        object_name = url.replace(endpoint, '')
        if object_name.startswith('/'):
            object_name = object_name[1:]
        return object_name

    def is_object_stale(
            self, container, name, filename, file_md5=None):
        """Check to see if an object matches the hashes of a file.
        :param container: Name of the container.
        :param name: Name of the object.
        :param filename: Path to the file.
        :param file_md5: Pre-calculated md5 of the file contents. Defaults to
            None which means calculate locally.
        """
        try:
            metadata = self.get_object_metadata(name, container)
        except (exceptions.NotFoundException, LocationParseError):
            self.log.debug(
                "swift stale check, no object: {container}/{name}".format(
                    container=container, name=name))
            return True

        if not file_md5:
            file_md5 = utils._get_file_hashes(filename)
        if file_md5 == metadata.etag.strip('\"'):
            self.log.debug(
                "swift object up to date: %(container)s/%(name)s",
                {'container': container, 'name': name})
            return False
        self.log.debug(
            "swift checksum mismatch: "
            " %(filename)s!=%(container)s/%(name)s",
            {'filename': filename, 'container': container, 'name': name})
        return True

    def get_object_segment_size(self, segment_size):
        """Get a segment size that will work given capabilities"""
        if segment_size is None:
            segment_size = DEFAULT_OBJECT_SEGMENT_SIZE
        min_segment_size = 0
        server_max_file_size = DEFAULT_MAX_FILE_SIZE

        if segment_size > server_max_file_size:
            return server_max_file_size
        if segment_size < min_segment_size:
            return min_segment_size
        return segment_size

    def copy_object(self):
        """Copy an object."""
        raise NotImplementedError

    def delete_object(self, obj, ignore_missing=True, container=None):
        """Delete an object

        :param obj: The value can be either the name of an object or a
                       :class:`~otcextensions.sdk.obs.v1.container.Container`
                       instance.
        :param container: The value can be the ID of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param bool ignore_missing: When set to ``False``
                    :class:`~openstack.exceptions.ResourceNotFound` will be
                    raised when the object does not exist.
                    When set to ``True``, no exception will be set when
                    attempting to delete a nonexistent server.

        :returns: ``None``
        """
        container_name = self._get_container_name(obj, container)
        endpoint = self.get_container_endpoint(container_name)
        self._delete(_obj.Object, obj, ignore_missing=ignore_missing,
                     container=container_name,
                     endpoint_override=endpoint,
                     requests_auth=self._get_req_auth(endpoint),
                     )

    def get_object_metadata(self, obj, container=None):
        """Get metadata for an object.

        :param obj: The value can be the name of an object or a
                    :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the ID of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.

        :returns: One :class:`~otcextensions.sdk.obs.v1.obj.Object`
        :raises: :class:`~openstack.exceptions.ResourceNotFound`
                 when no resource can be found.
        """
        container_name = self._get_container_name(obj, container)
        endpoint = self.get_container_endpoint(container_name)

        return self._head(_obj.Object, obj, container=container_name,
                          endpoint_override=endpoint,
                          requests_auth=self._get_req_auth(endpoint))

    def set_object_metadata(self, obj, container=None, **metadata):
        """Set metadata for an object.

        Note: This method will do an extra HEAD call.

        :param obj: The value can be the name of an object or a
                    :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param kwargs metadata: Key/value pairs to be set as metadata
                                on the container. Both custom and system
                                metadata can be set. Custom metadata are keys
                                and values defined by the user. System
                                metadata are keys defined by the Object Store
                                and values defined by the user. The system
                                metadata keys are:

                                - `content_type`
                                - `content_encoding`
                                - `content_disposition`
                                - `delete_after`
                                - `delete_at`
                                - `is_content_type_detected`
        """
        raise NotImplementedError
        container_name = self._get_container_name(obj, container)
        res = self._get_resource(_obj.Object, obj, container=container_name)
        res.set_metadata(self, metadata)
        return res

    def delete_object_metadata(self, obj, container=None, keys=None):
        """Delete metadata for an object.

        :param obj: The value can be the name of an object or a
                    :class:`~otcextensions.sdk.obs.v1.obj.Object` instance.
        :param container: The value can be the ID of a container or a
               :class:`~otcextensions.sdk.obs.v1.container.Container`
               instance.
        :param keys: The keys of metadata to be deleted.
        """
        raise NotImplementedError
        container_name = self._get_container_name(obj, container)
        res = self._get_resource(_obj.Object, obj, container=container_name)
        res.delete_metadata(self, keys)
        return res

    def _abort_multipart_upload(
            self, endpoint, upload_id
    ):
        """Aborts all multipart upload objects.
        :param endpoint: Container endpoint
        :param upload_id: ID of the multipart upload to be aborted.
        :return:
        """
        url = f'{endpoint}?uploadId={upload_id}'
        return self.delete(url, requests_auth=self._get_req_auth(endpoint))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest.mock import call

import mock
from openstackclient.tests.unit import utils as tests_utils
from osc_lib import exceptions

from otcextensions.osclient.modelartsv1.v1 import service
from otcextensions.tests.unit.osclient.modelartsv1.v1 import fakes

_COLUMNS = (
    "access_address",
    "additional_properties",
    "config",
    "description",
    "failed_times",
    "id",
    "infer_type",
    "invocation_times",
    "is_free",
    "is_shared",
    "name",
    "operation_time",
    "owner",
    "progress",
    "project",
    "publish_at",
    "service_id",
    "service_name",
    "shared_count",
    "status",
    "tenant",
    "transition_at",
    "update_time",
    "workspace_id",
)


class TestListServices(fakes.TestModelartsv1):
    objects = fakes.FakeService.create_multiple(3)

    column_list_headers = (
        "Service Id",
        "Service Name",
        "Infer Type",
        "Status",
    )

    data = []

    for s in objects:
        data.append((s.service_id, s.service_name, s.infer_type, s.status))

    def setUp(self):
        super(TestListServices, self).setUp()

        self.cmd = service.ListServices(self.app, None)

        self.client.services = mock.Mock()
        self.client.api_mock = self.client.services

    def test_list(self):
        arglist = []

        verifylist = []

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Set the response
        self.client.api_mock.side_effect = [self.objects]

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)

        self.client.api_mock.assert_called_with()

        self.assertEqual(self.column_list_headers, columns)
        self.assertEqual(self.data, list(data))

    def test_list_args(self):
        arglist = [
            "--cluster-id",
            "4",
            "--infer-type",
            "real-time",
            "--limit",
            "7",
            "--model-id",
            "3",
            "--offset",
            "6",
            "--order",
            "asc",
            "--service-id",
            "1",
            "--name",
            "2",
            "--sort-by",
            "service_name",
            "--status",
            "running",
            "--workspace-id",
            "5",
        ]

        verifylist = [
            ("cluster_id", "4"),
            ("infer_type", "real-time"),
            ("limit", 7),
            ("model_id", "3"),
            ("offset", 6),
            ("order", "asc"),
            ("service_id", "1"),
            ("name", "2"),
            ("sort_by", "service_name"),
            ("status", "running"),
            ("workspace_id", "5"),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Set the response
        self.client.api_mock.side_effect = [self.objects]

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)

        self.client.api_mock.assert_called_with(
            cluster_id="4",
            infer_type="real-time",
            limit=7,
            model_id="3",
            name="2",
            offset=6,
            order="asc",
            service_id="1",
            sort_by="service_name",
            status="running",
            workspace_id="5",
        )


class TestCreateService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_service, columns, service._formatters)

    default_timeout = 1200

    def setUp(self):
        super(TestCreateService, self).setUp()

        self.cmd = service.CreateService(self.app, None)

        self.client.create_service = mock.Mock(return_value=self._service)
        self.client.get_service = mock.Mock(return_value=self._service)
        self.client.wait_for_service = mock.Mock(return_value=True)

    def test_create_batch(self):
        arglist = [
            "test-service",
            "--infer-type",
            "batch",
            "--router-id",
            "1",
            "--network-id",
            "2",
            "--security-group-id",
            "3",
            "--model-id",
            "4",
            "--specification",
            "5",
            "--instance-count",
            "6",
            "--src-path",
            "7",
            "--dest-path",
            "8",
            "--req-uri",
            "9",
            "--mapping-type",
            "csv",
            "--env",
            "VAR1=value1",
            "--env",
            "VAR2=value2",
            "--wait",
        ]
        verifylist = [
            ("name", "test-service"),
            ("infer_type", "batch"),
            ("vpc_id", "1"),
            ("subnet_network_id", "2"),
            ("security_group_id", "3"),
            ("model_id", "4"),
            ("specification", "5"),
            ("instance_count", 6),
            ("src_path", "7"),
            ("dest_path", "8"),
            ("req_uri", "9"),
            ("mapping_type", "csv"),
            ("envs", {"VAR1": "value1", "VAR2": "value2"}),
            ("wait", True),
        ]
        # Verify cm is triggereg with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            "service_name": "test-service",
            "infer_type": "batch",
            "config": [
                {
                    "model_id": "4",
                    "specification": "5",
                    "instance_count": 6,
                    "envs": {"VAR1": "value1", "VAR2": "value2"},
                    "src_path": "7",
                    "dest_path": "8",
                    "req_uri": "9",
                    "mapping_type": "csv",
                }
            ],
            "vpc_id": "1",
            "subnet_network_id": "2",
            "security_group_id": "3",
        }
        self.client.create_service.assert_called_with(**attrs)
        self.client.wait_for_service.assert_called_with(
            self._service.id, self.default_timeout
        )
        self.client.get_service.assert_called_with(self._service.id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_realtime(self):
        arglist = [
            "test-service",
            "--infer-type",
            "real-time",
            "--router-id",
            "1",
            "--network-id",
            "2",
            "--security-group-id",
            "3",
            "--model-id",
            "4",
            "--specification",
            "5",
            "--instance-count",
            "6",
            "--weight",
            "7",
            "--env",
            "VAR1=value1",
            "--env",
            "VAR2=value2",
            "--wait",
        ]
        verifylist = [
            ("name", "test-service"),
            ("infer_type", "real-time"),
            ("vpc_id", "1"),
            ("subnet_network_id", "2"),
            ("security_group_id", "3"),
            ("model_id", "4"),
            ("specification", "5"),
            ("instance_count", 6),
            ("weight", 7),
            ("envs", {"VAR1": "value1", "VAR2": "value2"}),
            ("wait", True),
        ]
        # Verify cm is triggereg with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            "service_name": "test-service",
            "infer_type": "real-time",
            "config": [
                {
                    "model_id": "4",
                    "specification": "5",
                    "instance_count": 6,
                    "envs": {"VAR1": "value1", "VAR2": "value2"},
                    "weight": 7,
                }
            ],
            "vpc_id": "1",
            "subnet_network_id": "2",
            "security_group_id": "3",
        }
        self.client.create_service.assert_called_with(**attrs)
        self.client.wait_for_service.assert_called_with(
            self._service.id, self.default_timeout
        )
        self.client.get_service.assert_called_with(self._service.id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestShowService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_service, columns, service._formatters)

    def setUp(self):
        super(TestShowService, self).setUp()

        self.cmd = service.ShowService(self.app, None)

        self.client.find_service = mock.Mock(return_value=self._service)
        self.client.get_service = mock.Mock(return_value=self._service)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Testing that a call without the required argument will fail and
        # throw a "ParserExecption"
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show(self):
        arglist = [
            self._service.name,
        ]

        verifylist = [("service", self._service.name)]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        self.client.find_service.assert_called_with(
            self._service.name, ignore_missing=False
        )
        self.client.get_service.assert_called_with(self._service.id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_show_non_existent(self):
        arglist = ["nonexisting_service"]

        verifylist = [("service", "nonexisting_service")]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = exceptions.CommandError("Resource Not Found")
        self.client.find_service = mock.Mock(side_effect=find_mock_result)

        # Trigger the action
        try:
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.assertEqual("Resource Not Found", str(e))
        self.client.find_service.assert_called_with(
            "nonexisting_service", ignore_missing=False
        )


class TestUpdateService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_service, columns, service._formatters)

    default_timeout = 1200

    def setUp(self):
        super(TestUpdateService, self).setUp()

        self.cmd = service.UpdateService(self.app, None)

        self.client.find_service = mock.Mock(return_value=self._service)
        self.client.get_service = mock.Mock(return_value=self._service)
        self.client.update_service = mock.Mock(return_value=self._service)
        self.client.wait_for_service = mock.Mock(return_value=True)

    def test_update(self):
        arglist = [
            self._service.name,
            "--schedule",
            "type=1,time_unit=2,duration=3",
            "--additional-property",
            "key1=value1",
            "--additional-property",
            "key2=value2",
            "--model-id",
            "model-id",
            "--weight",
            "2",
            "--specification",
            "3",
            "--instance-count",
            "1",
            "--env",
            "key1=value1",
            "--custom-spec",
            "cpu=2,memory=8",
            "--src-type",
            "5",
            "--src-path",
            "6",
            "--dest-path",
            "7",
            "--req-uri",
            "8",
            "--mapping-type",
            "file",
            "--mapping-rule",
            "9",
            "--wait",
        ]
        verifylist = [
            ("service", self._service.name),
            (
                "schedule",
                [{"type": "1", "time_unit": "2", "duration": "3"}],
            ),
            (
                "additional_properties",
                {"key1": "value1", "key2": "value2"},
            ),
            ("model_id", "model-id"),
            ("weight", 2),
            ("specification", "3"),
            ("instance_count", 1),
            ("envs", {"key1": "value1"}),
            ("custom_spec", [{"cpu": "2", "memory": "8"}]),
            ("src_type", "5"),
            ("src_path", "6"),
            ("dest_path", "7"),
            ("req_uri", "8"),
            ("mapping_type", "file"),
            ("mapping_rule", "9"),
            ("wait", True),
            ("timeout", self.default_timeout),
        ]
        # Verify cm is triggereg with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            "schedule": [
                {
                    "type": "1",
                    "time_unit": "2",
                    "duration": "3",
                }
            ],
            "additional_properties": {
                "key1": "value1",
                "key2": "value2",
            },
            "config": [
                {
                    "model_id": "model-id",
                    "specification": "3",
                    "instance_count": 1,
                    "envs": {
                        "key1": "value1",
                    },
                    "weight": 2,
                    "custom_spec": {
                        "cpu": 2.0,
                        "memory": 8,
                    },
                }
            ],
        }

        self.client.update_service.assert_called_with(
            self._service.id, **attrs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestStartService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_service, columns)

    def setUp(self):
        super(TestStartService, self).setUp()

        self.cmd = service.StartService(self.app, None)

        self.client.find_service = mock.Mock(return_value=self._service)
        self.client.start_service = mock.Mock(return_value=None)

    def test_start(self):
        arglist = [
            self._service.id,
        ]

        verifylist = [
            ("service", self._service.id),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        self.cmd.take_action(parsed_args)
        self.client.find_service.assert_called_with(
            self._service.id, ignore_missing=False
        )
        self.client.start_service.assert_called_with(self._service)


class TestStopService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_service, columns)

    def setUp(self):
        super(TestStopService, self).setUp()

        self.cmd = service.StopService(self.app, None)

        self.client.find_service = mock.Mock(return_value=self._service)
        self.client.stop_service = mock.Mock(return_value=None)

    def test_start(self):
        arglist = [
            self._service.id,
        ]

        verifylist = [
            ("service", self._service.id),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        self.cmd.take_action(parsed_args)
        self.client.find_service.assert_called_with(
            self._service.id, ignore_missing=False
        )
        self.client.stop_service.assert_called_with(self._service)


class TestDeleteService(fakes.TestModelartsv1):
    _service = fakes.FakeService.create_multiple(2)

    def setUp(self):
        super(TestDeleteService, self).setUp()

        self.client.find_service = mock.Mock(return_value=self._service[0])
        self.client.delete_service = mock.Mock(return_value=None)

        # Get the command object to test
        self.cmd = service.DeleteService(self.app, None)

    def test_delete(self):
        arglist = [
            self._service[0].name,
        ]

        verifylist = [
            ("service", arglist),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        result = self.cmd.take_action(parsed_args)
        self.client.find_service.assert_called_with(
            self._service[0].name, ignore_missing=False
        )
        self.client.delete_service.assert_called_with(self._service[0].id)
        self.assertIsNone(result)

    def test_multiple_delete(self):
        arglist = []

        for ma_service in self._service:
            arglist.append(ma_service.name)

        verifylist = [("service", arglist)]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_results = self._service
        self.client.find_service = mock.Mock(side_effect=find_mock_results)

        # Trigger the action
        result = self.cmd.take_action(parsed_args)

        find_calls = []
        delete_calls = []
        for ma_service in self._service:
            find_calls.append(call(ma_service.name, ignore_missing=False))
            delete_calls.append(call(ma_service.id))
        self.client.find_service.assert_has_calls(find_calls)
        self.client.delete_service.assert_has_calls(delete_calls)
        self.assertIsNone(result)

    def test_multiple_delete_with_exception(self):
        arglist = [
            self._service[0].id,
            "unexist_ma_service",
        ]
        verifylist = [("service", arglist)]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_results = [
            self._service[0],
            exceptions.CommandError,
        ]
        self.client.find_service = mock.Mock(side_effect=find_mock_results)

        # Trigger the action
        try:
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.assertEqual("1 of 2 Service(s) failed to delete.", str(e))

        self.client.delete_service.assert_any_call(self._service[0].id)



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
'''
Create CloudEye alarm rule
'''
import openstack

openstack.enable_logging(True)
conn = openstack.connect(cloud='otc')


attrs = {
    "alarm_name": "alarm-test",
    "alarm_description": "Test Alarm description",
    "metric": {
        "namespace": "SYS.ECS",
        "dimensions": [
            {
                "name": "instance_id",
                "value": "33328f02-3814-422e-b688-bfdba93d4051"
            },
            {
                "name": "instance_id",
                "value": "04ab9572-8c9c-41b6-bcc8-51068463b123"
            }
        ],
        "metric_name": "network_outgoing"
    },
    "condition": {
        "period": 300,
        "filter": "average",
        "comparison_operator": ">=",
        "value": 6,
        "unit": "B/s",
        "count": 1
    },
    "alarm_enabled": True,
    "alarm_action_enabled": True,
    "alarm_level": 2,
    "ok_actions": [
        {
            "type": "notification",
            "notificationList": [
                "urn:smn:region:68438a86d98e427e907e0097b7e35d48:sd",
                "urn:smn:eu-de:16d53a84a13b49529d2e2c3646691222:Error"]
        }
    ],
    "alarm_actions": [
        {
            "type": "notification",
            "notificationList": [
                "urn:smn:region:68438a86d98e427e907e0097b7e35d48:sd",
                "urn:smn:eu-de:16d53a84a13b49529d2e2c3646691222:Error"]
        }
    ]
}


alarm = conn.ces.create_alarm(**attrs)
print(alarm)

# OSC command
'''
openstack --os-cloud otc ces alarm create --description "Test Alarm" \
--namespace SYS.ECS \
--dimension name=instance_id,value=33328f02-3814-422e-b688-bfdba93d4123 \
--dimension name=instance_id,value=33328f02-3814-422e-b688-bfdba93d4052 \
--metric-name "network_outgoing" --period '300' --filter average \
--comparison-operator '>=' --value '6' --unit 'B/s' --count '1' \
--enabled True --action-enabled True --level 2 --ok-action-type notification \
--ok-action-notification-list \
'urn:smn:region:68438a86d98e427e907e0097b7e35d48:sd' \
--ok-action-notification-list \
'urn:smn:eu-de:16d53a84a13b49529d2e2c3646691222:Error' \
--alarm-action-type notification --alarm-action-notification-list \
'urn:smn:region:68438a86d98e427e907e0097b7e35d48:sd' \
--alarm-action-notification-list \
'urn:smn:eu-de:16d53a84a13b49529d2e2c3646691222:Error' alarm-test
'''



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from urllib import parse

from openstack import proxy
from openstack import resource

from otcextensions.sdk.cce.v3 import cluster as _cluster
from otcextensions.sdk.cce.v3 import cluster_cert as _cluster_cert
from otcextensions.sdk.cce.v3 import cluster_node as _cluster_node
from otcextensions.sdk.cce.v3 import job as _job
from otcextensions.sdk.cce.v3 import node_pool as _node_pool


class Proxy(proxy.Proxy):

    skip_discovery = True

    def _extract_name(self, url, service_type=None, project_id=None):
        url_path = parse.urlparse(url).path.strip()
        # Remove / from the beginning to keep the list indexes of interesting
        # things consistent
        if url_path.startswith('/'):
            url_path = url_path[1:]

        # Split url into parts and exclude potential project_id in some urls
        url_parts = [
            x for x in url_path.split('/') if (
                x != project_id
                and (
                    not project_id
                    or (project_id and x != project_id)
                ))
        ]
        # Strip leading version piece so that
        # GET /api/v3/projects/xxx
        # returns []
        if (
            len(url_parts) >= 3
            and url_parts[0] == 'api' and url_parts[2] == 'projects'
        ):
            url_parts = url_parts[3:]

        name_parts = self._extract_name_consume_url_parts(url_parts)

        if not name_parts:
            name_parts = ['discovery']

        # Strip out anything that's empty or None
        return [part for part in name_parts if part]

    # ======== Cluster ========
    def clusters(self):
        """List all Clusters.

        :returns: a generator of
            (:class:`~otcextensions.sdk.cce.v3.cluster.Cluster`) instances
        """
        return self._list(_cluster.Cluster, paginated=False)

    def get_cluster(self, cluster):
        """Get the cluster by UUID.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        """
        return self._get(
            _cluster.Cluster, cluster,
        )

    def create_cluster(self, **attrs):
        """Create a cluster from attributes.

        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`,
            comprised of the properties on the Cluster class.

        :returns: The results of cluster creation
        :rtype: :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        """
        return self._create(
            _cluster.Cluster, prepend_key=False, **attrs
        )

    def find_cluster(self, name_or_id, ignore_missing=True):
        """Find a single cluster.

        :param name_or_id: The name or ID of a cluster
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be raised
            when the group does not exist.
            When set to ``True``, no exception will be set when attempting
            to delete a nonexistent group.

        :returns: ``None``
        """
        return self._find(
            _cluster.Cluster, name_or_id,
            ignore_missing=ignore_missing,
        )

    def delete_cluster(self, cluster, ignore_missing=True):
        """Delete a cluster.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
            the group does not exist.
            When set to ``True``, no exception will be set when attempting to
            delete a nonexistent cluster.
        """
        return self._delete(
            _cluster.Cluster, cluster, ignore_missing=ignore_missing,
        )

    def get_cluster_certificates(self, cluster):
        """Get the certificates of a cluster.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.cluster_cert.ClusterCertificates`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._get(
            _cluster_cert.ClusterCertificate, cluster_id=cluster.id,
            requires_id=False
        )

    def wait_for_cluster(self, cluster, status='Available', failures=None,
                         interval=2, wait=960):
        """Wait for a server to be in a particular status.

        :param cluster:
            The :class:`~otcextensions.sdk.cce.v3.cluster.Cluster` to wait on
            to reach the specified status.
        :param status: Desired status.
        :param failures:
            Statuses that would be interpreted as failures.
        :type failures: :py:class:`list`
        :param int interval:
            Number of seconds to wait before to consecutive checks.
            Default to 2.
        :param int wait:
            Maximum number of seconds to wait before the change.
            Default to 960.
        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
                 to the desired status failed to occur in specified seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
                 has transited to one of the failure statuses.
        """
        failures = ['ERROR'] if failures is None else failures
        return resource.wait_for_status(
            self, cluster, status, failures, interval, wait,
            attribute='status.status')

    # ======== Cluster Nodes ========
    def cluster_nodes(self, cluster):
        """List all Cluster nodes.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.

        :returns: a generator of
            (:class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`)
            instances
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._list(
            _cluster_node.ClusterNode, cluster_id=cluster.id,
            paginated=False
        )

    def get_cluster_node(self, cluster, node_id):
        """Get the cluster node by it's UUID.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        :param node_id: Cluster node id to be fetched

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._get(
            _cluster_node.ClusterNode,
            node_id,
            cluster_id=cluster.id,
        )

    def find_cluster_node(self, cluster, node):
        """Find the cluster node by it's UUID or name.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        :param node: Cluster node id or name to be fetched

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._find(
            _cluster_node.ClusterNode,
            node,
            cluster_id=cluster.id,
        )

    def delete_cluster_node(self, cluster, node, ignore_missing=True):
        """Delete nodes from the cluster.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.
        :param node: The value can be the ID of a cluster node
             or a :class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`
             instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
            the node does not exist.
            When set to ``True``, no exception will be set when attempting to
            delete a nonexistent cluster node.
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._delete(
            _cluster_node.ClusterNode,
            node,
            ignore_missing=ignore_missing,
            cluster_id=cluster.id,
        )

    def create_cluster_node(self, cluster, **attrs):
        """Add a new node to the cluster.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.
        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`,
            comprised of the properties on the ClusterNode class.
        :returns: The results of config creation
        :rtype: :class:`~otcextensions.sdk.cce.v3.cluster_node.ClusterNode`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._create(
            _cluster_node.ClusterNode,
            cluster_id=cluster.id,
            **attrs
        )

    # ======== Node Pools ========
    def node_pools(self, cluster):
        """List all CCE node pools.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.

        :returns: a generator of
            (:class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`)
            instances
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._list(
            _node_pool.NodePool, cluster_id=cluster.id,
            paginated=False
        )

    def get_node_pool(self, cluster, node_pool_id):
        """Get the CCE node pool by it's UUID.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        :param node_pool_id: Node pool id to be fetched

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._get(
            _node_pool.NodePool,
            node_pool_id,
            cluster_id=cluster.id,
        )

    def find_node_pool(self, cluster, node_pool):
        """Find the cluster node by it's UUID or name.

        :param cluster: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
        :param node_pool: Node pool id or name to be fetched

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._find(
            _node_pool.NodePool,
            node_pool,
            cluster_id=cluster.id,
        )

    def delete_node_pool(self, cluster, node_pool, ignore_missing=True):
        """Delete node pool from the cluster.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.
        :param node_pool: The value can be the ID of a CCE node pool
             or a :class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`
             instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
            the node does not exist.
            When set to ``True``, no exception will be set when attempting to
            delete a nonexistent cluster node pool.
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._delete(
            _node_pool.NodePool,
            node_pool,
            ignore_missing=ignore_missing,
            cluster_id=cluster.id,
        )

    def create_node_pool(self, cluster, **attrs):
        """Add a new node pool to the cluster.

        :param cluster: The value can be the ID of a cluster
             or a :class:`~otcextensions.sdk.cce.v3.cluster.Cluster`
             instance.
        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`,
            comprised of the properties on the NodePool class.
        :returns: The results of config creation
        :rtype: :class:`~otcextensions.sdk.cce.v3.node_pool.NodePool`
        """
        cluster = self._get_resource(_cluster.Cluster, cluster)
        return self._create(
            _node_pool.NodePool,
            cluster_id=cluster.id,
            **attrs
        )

    # ======== Job Operations ========
    def get_job(self, job):
        """Get the job by UUID.

        :param job: key id or an instance of
            :class:`~otcextensions.sdk.cce.v3.job.Job`

        :returns: instance of
            :class:`~otcextensions.sdk.cce.v3.job.Job`
        """
        return self._get(
            _job.Job, job,
        )

    def wait_for_job(self, job_id, status='success',
                     failures=None, interval=5, wait=3600,
                     attribute='status.status'):
        failures = ['FAILED'] if failures is None else failures
        job = self.get_job(job_id)
        return resource.wait_for_status(
            self, job, status, failures, interval, wait,
            attribute='status.status')

    # ======== Project cleanup ========
    def _get_cleanup_dependencies(self):
        return {
            'cce': {
                'before': ['compute', 'network']
            }
        }

    def _service_cleanup(
        self, dry_run=True, client_status_queue=None,
        identified_resources=None,
        filters=None, resource_evaluation_fn=None,
        skip_resources=None
    ):
        for obj in self.clusters():
            self._service_cleanup_del_res(
                self.delete_cluster,
                obj,
                dry_run=dry_run,
                client_status_queue=client_status_queue,
                identified_resources=identified_resources,
                filters=filters,
                resource_evaluation_fn=resource_evaluation_fn)



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Set CSS Cluster Snapshot Basic Configuration
"""

import openstack

openstack.enable_logging(True)
conn = openstack.connect(cloud='otc')

cluster_id = 'cluster-uuid'

# Set Cluster Snapshot Configuration automatically.
conn.css.set_snapshot_configuration(cluster_id, auto_configure=True)

# Set Cluster Snapshot Configuration with custom bucket & agency.
attrs = {
    'bucket': 'css-backup-1626212749424',
    'agency': 'css_obs_agency',
    'backup_path': 'css/repository/test-css',
}
conn.css.set_snapshot_configuration(cluster_id, **attrs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest.mock import call

import mock
from openstackclient.tests.unit import utils as tests_utils
from osc_lib import exceptions

from otcextensions.common import cli_utils
from otcextensions.osclient.modelartsv1.v1 import devenv
from otcextensions.tests.unit.osclient.modelartsv1.v1 import fakes

# from osc_lib.cli import format_columns


_COLUMNS = (
    "created_at",
    "description",
    "flavor",
    "flavor_details",
    "id",
    "name",
    "profile",
    "spec",
    "status",
    "updated_at",
    "user",
)


class TestListDevenvInstances(fakes.TestModelartsv1):
    objects = fakes.FakeDevenv.create_multiple(3)

    column_list_headers = ("ID", "Name", "status", "Created At")

    columns = ("id", "name", "status", "created_at")

    data = []

    for s in objects:
        data.append(
            (
                s.id,
                s.name,
                s.status,
                cli_utils.UnixTimestampFormatter(s.created_at),
            )
        )

    def setUp(self):
        super(TestListDevenvInstances, self).setUp()

        self.cmd = devenv.ListDevenvInstances(self.app, None)

        self.client.devenv_instances = mock.Mock()
        self.client.api_mock = self.client.devenv_instances

    def test_list(self):
        arglist = []

        verifylist = [
            ("dev_env_type", "Notebook"),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Set the response
        self.client.api_mock.side_effect = [self.objects]

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)

        self.client.api_mock.assert_called_with(dev_env_type="Notebook")

        self.assertEqual(self.column_list_headers, columns)
        self.assertEqual(self.data, list(data))

    def test_list_args(self):
        arglist = [
            "--dev-env-type",
            "1",
            "--provision-type",
            "2",
            "--status",
            "running",
            "--sort-by",
            "name",
            "--order",
            "asc",
            "--offset",
            "6",
            "--limit",
            "7",
            "--workspace-id",
            "8",
            "--ai-project-id",
            "9",
            "--pool-id",
            "10",
            "--show-self",
        ]

        verifylist = [
            ("dev_env_type", "1"),
            ("provision_type", "2"),
            ("status", "RUNNING"),
            ("sort_by", "name"),
            ("order", "asc"),
            ("offset", 6),
            ("limit", 7),
            ("workspace_id", "8"),
            ("ai_project", "9"),
            ("pool_id", "10"),
            ("show_self", True),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Set the response
        self.client.api_mock.side_effect = [self.objects]

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)

        self.client.api_mock.assert_called_with(
            ai_project="9",
            dev_env_type="1",
            limit=7,
            offset=6,
            order="asc",
            pool_id="10",
            provision_type="2",
            show_self=True,
            sort_by="name",
            status="RUNNING",
            workspace_id="8",
        )


class TestCreateDevenvInstance(fakes.TestModelartsv1):
    _data = fakes.FakeDevenv.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_data, columns, devenv._formatters)
    default_timeout = 1800

    def setUp(self):
        super(TestCreateDevenvInstance, self).setUp()

        self.cmd = devenv.CreateDevenvInstance(self.app, None)

        self.client.create_devenv_instance = mock.Mock(return_value=self._data)

    def test_create(self):
        arglist = [
            "test-devenv",
            "--profile-id",
            "1",
            "--description",
            "2",
            "--flavor",
            "3",
            "--ai-project-id",
            "4",
            "--workspace-id",
            "5",
            "--storage-type",
            "6",
            "--storage-path",
            "7",
            "--extended-storage-type",
            "8",
            "--extended-storage-path",
            "9",
            "--autostop-duration",
            "10",
            "--autostop-prompt",
            False,
            "--pool-id",
            "11",
            "--pool-type",
            "12",
            "--pool-name",
            "13",
            "--annotation",
            "key1=value1",
            "--annotation",
            "key2=value2",
        ]
        verifylist = [
            ("name", "test-devenv"),
            ("profile_id", 1),
            ("description", "2"),
            ("flavor", "3"),
            ("ai_project_id", "4"),
            ("workspace_id", "5"),
            ("storage_type", "6"),
            ("storage_path", "7"),
            ("extended_storage_type", "8"),
            ("extended_storage_path", "9"),
            ("autostop_duration", 10),
            ("autostop_prompt", False),
            ("pool_id", "11"),
            ("pool_type", "12"),
            ("pool_name", "13"),
            ("annotation", {"key1": "value1", "key2": "value2"}),
        ]
        # Verify cm is triggereg with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            "name": "test-devenv",
            "flavor": "3",
            "profile_id": "Multi-Engine 1.0 (python3)-cpu",
            "spec": {
                "storage": {"location": {"path": "7"}, "type": "6"},
                "auto_stop": {"enable": True, "duration": 10},
                "extended_storage": {
                    "location": {"path": "9"},
                    "type": "8",
                },
            },
            "description": "2",
            "workspace": {"id": "5"},
            "ai_project": {"id": "4"},
            "pool": {"name": "13"},
            "annotations": {"key1": "value1", "key2": "value2"},
        }
        self.client.create_devenv_instance.assert_called_with(**attrs)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestUpdateDevenvInstance(fakes.TestModelartsv1):
    _data = fakes.FakeDevenv.create_one()

    columns = _COLUMNS

    data = fakes.gen_data(_data, columns, devenv._formatters)

    default_timeout = 1800

    def setUp(self):
        super(TestUpdateDevenvInstance, self).setUp()

        self.cmd = devenv.UpdateDevenvInstance(self.app, None)

        self.client.update_devenv_instance = mock.Mock(return_value=self._data)

    def test_update(self):
        arglist = [
            "devenv-instance-id",
            "--description",
            "New Description",
            "--auto-stop",
            "enable",
            "--duration",
            "3600",
            "--prompt",
            "disable",
        ]
        verifylist = [
            ("instance", "devenv-instance-id"),
            ("description", "New Description"),
            ("auto_stop", "enable"),
            ("duration", 3600),
            ("prompt", "disable"),
        ]
        # Verify cm is triggereg with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        attrs = {
            "auto_stop": {
                "enable": True,
                "prompt": False,
                "duration": 3600,
            },
            "description": "New Description",
        }
        self.client.update_devenv_instance.assert_called_with(
            "devenv-instance-id", **attrs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestStartDevenvInstance(fakes.TestModelartsv1):
    _devenv = fakes.FakeDevenv.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_devenv, columns, devenv._formatters)

    def setUp(self):
        super(TestStartDevenvInstance, self).setUp()

        self.cmd = devenv.StartDevenvInstance(self.app, None)

        self.client.find_devenv_instance = mock.Mock(return_value=self._devenv)
        self.client.start_devenv_instance = mock.Mock(
            return_value=self._devenv
        )

    def test_start(self):
        arglist = [
            self._devenv.id,
        ]

        verifylist = [
            ("instance", self._devenv.id),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        self.cmd.take_action(parsed_args)
        self.client.find_devenv_instance.assert_called_with(
            self._devenv.id, ignore_missing=False
        )
        self.client.start_devenv_instance.assert_called_with(self._devenv.id)


class TestStopDevenvInstance(fakes.TestModelartsv1):
    _devenv = fakes.FakeDevenv.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_devenv, columns, devenv._formatters)

    def setUp(self):
        super(TestStopDevenvInstance, self).setUp()

        self.cmd = devenv.StopDevenvInstance(self.app, None)

        self.client.find_devenv_instance = mock.Mock(return_value=self._devenv)
        self.client.stop_devenv_instance = mock.Mock(return_value=self._devenv)

    def test_start(self):
        arglist = [
            self._devenv.id,
        ]

        verifylist = [
            ("instance", self._devenv.id),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        self.cmd.take_action(parsed_args)
        self.client.find_devenv_instance.assert_called_with(
            self._devenv.id, ignore_missing=False
        )
        self.client.stop_devenv_instance.assert_called_with(self._devenv.id)


class TestShowDevenvInstance(fakes.TestModelartsv1):
    _devenv = fakes.FakeDevenv.create_one()
    columns = _COLUMNS
    data = fakes.gen_data(_devenv, columns, devenv._formatters)

    def setUp(self):
        super(TestShowDevenvInstance, self).setUp()

        self.cmd = devenv.ShowDevenvInstance(self.app, None)

        self.client.find_devenv_instance = mock.Mock(return_value=self._devenv)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Testing that a call without the required argument will fail and
        # throw a "ParserExecption"
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show(self):
        arglist = [
            self._devenv.id,
        ]

        verifylist = [
            ("instance", self._devenv.id),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        columns, data = self.cmd.take_action(parsed_args)
        self.client.find_devenv_instance.assert_called_with(
            self._devenv.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_show_non_existent(self):
        arglist = [
            "unexist_devenv_instance",
        ]

        verifylist = [
            ("instance", "unexist_devenv_instance"),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = exceptions.CommandError("Resource Not Found")
        self.client.find_devenv_instance = mock.Mock(
            side_effect=find_mock_result
        )

        # Trigger the action
        try:
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.assertEqual("Resource Not Found", str(e))
        self.client.find_devenv_instance.assert_called_with(
            "unexist_devenv_instance", ignore_missing=False
        )


class TestDeleteDevenvInstance(fakes.TestModelartsv1):
    _devenv = fakes.FakeDevenv.create_multiple(2)

    def setUp(self):
        super(TestDeleteDevenvInstance, self).setUp()

        self.client.find_devenv_instance = mock.Mock(
            return_value=self._devenv[0]
        )
        self.client.delete_devenv_instance = mock.Mock(return_value=None)

        # Get the command object to test
        self.cmd = devenv.DeleteDevenvInstance(self.app, None)

    def test_delete(self):
        arglist = [
            self._devenv[0].name,
        ]

        verifylist = [
            ("instance", arglist),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Trigger the action
        result = self.cmd.take_action(parsed_args)
        self.client.find_devenv_instance.assert_called_with(
            self._devenv[0].name, ignore_missing=False
        )
        self.client.delete_devenv_instance.assert_called_with(
            self._devenv[0].id
        )
        self.assertIsNone(result)

    def test_multiple_delete(self):
        arglist = []

        for instance in self._devenv:
            arglist.append(instance.name)

        verifylist = [
            ("instance", arglist),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_results = self._devenv
        self.client.find_devenv_instance = mock.Mock(
            side_effect=find_mock_results
        )

        # Trigger the action
        result = self.cmd.take_action(parsed_args)

        find_calls = []
        delete_calls = []
        for instance in self._devenv:
            find_calls.append(call(instance.name, ignore_missing=False))
            delete_calls.append(call(instance.id))
        self.client.find_devenv_instance.assert_has_calls(find_calls)
        self.client.delete_devenv_instance.assert_has_calls(delete_calls)
        self.assertIsNone(result)

    def test_multiple_delete_with_exception(self):
        arglist = [
            self._devenv[0].id,
            "unexist_devenv_instance",
        ]
        verifylist = [
            ("instance", arglist),
        ]

        # Verify cm is triggered with default parameters
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_results = [self._devenv[0], exceptions.CommandError]
        self.client.find_devenv_instance = mock.Mock(
            side_effect=find_mock_results
        )

        # Trigger the action
        try:
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.assertEqual(
                "1 of 2 Devenv Instance(s) failed to delete.", str(e)
            )

        self.client.delete_devenv_instance.assert_any_call(self._devenv[0].id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from openstack import exceptions

from otcextensions.tests.functional import base


class TestElb(base.BaseFunctionalTest):
    uuid_v4 = uuid.uuid4().hex[:8]
    network = None
    load_balancer = None
    listener = None

    def setUp(self):
        super(TestElb, self).setUp()
        self.client = self.conn.elb
        self.network_client = self.conn.network

    def create_load_balancer(
            self,
            name='sdk-elb-test-lb-' + uuid_v4,
            description='test',
            **kwargs
    ):
        attrs = {
            'name': name,
            'description': description,
            'vip_subnet_id': TestElb.network['subnet_id'],
            **kwargs
        }
        if TestElb.network and not TestElb.load_balancer:
            TestElb.load_balancer = self.client.create_load_balancer(**attrs)

    def create_listener(
            self,
            name='sdk-elb-test-lis-' + uuid_v4,
            protocol_port=80,
            protocol='TCP',
            **kwargs
    ):
        attrs = {
            'protocol_port': protocol_port,
            'protocol': protocol,
            'name': name,
            **kwargs
        }
        if not TestElb.load_balancer:
            raise exceptions.SDKException
        attrs['loadbalancer_id'] = TestElb.load_balancer.id
        if TestElb.network and TestElb.load_balancer \
                and not TestElb.listener:
            TestElb.listener = self.client.create_listener(**attrs)

    def create_network(
            self,
            cidr='192.168.0.0/16',
            ip_version=4,
            router_name='sdk-elb-test-router-',
            net_name='sdk-elb-test-net-',
            subnet_name='sdk-elb-test-subnet-'
    ):
        uuid_v4 = uuid.uuid4().hex[:8]
        router_name += uuid_v4
        net_name += uuid_v4
        subnet_name += uuid_v4
        if not TestElb.network:
            network = self.conn.network.create_network(name=net_name)
            self.assertEqual(net_name, network.name)
            net_id = network.id
            subnet = self.conn.network.create_subnet(
                name=subnet_name,
                ip_version=ip_version,
                network_id=net_id,
                cidr=cidr
            )
            self.assertEqual(subnet_name, subnet.name)
            subnet_id = subnet.id

            router = self.conn.network.create_router(name=router_name)
            self.assertEqual(router_name, router.name)
            router_id = router.id
            interface = router.add_interface(
                self.conn.network,
                subnet_id=subnet_id
            )
            self.assertEqual(interface['subnet_id'], subnet_id)
            self.assertIn('port_id', interface)
            TestElb.network = {
                'router_id': router_id,
                'subnet_id': subnet_id,
                'network_id': net_id
            }
        return

    def destroy_network(self, params: dict):
        router_id = params.get('router_id')
        subnet_id = params.get('subnet_id')
        network_id = params.get('network_id')
        router = self.conn.network.get_router(router_id)

        interface = router.remove_interface(
            self.conn.network,
            subnet_id=subnet_id
        )
        self.assertEqual(interface['subnet_id'], subnet_id)
        self.assertIn('port_id', interface)
        sot = self.conn.network.delete_router(
            router_id,
            ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.conn.network.delete_subnet(
            subnet_id,
            ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.conn.network.delete_network(
            network_id,
            ignore_missing=False
        )
        self.assertIsNone(sot)
        TestElb.network = None



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
'''DMS Instance AZ action implementations'''
from osc_lib import utils
from osc_lib.command import command

from otcextensions.i18n import _


class ListAZ(command.Lister):
    _description = _('List Availability zones')
    columns = ('ID', 'name', 'code', 'port', 'has_available_resources')

    def get_parser(self, prog_name):
        parser = super(ListAZ, self).get_parser(prog_name)

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.dms

        data = client.availability_zones()

        table = (self.columns,
                 (utils.get_item_properties(
                     s, self.columns,
                 ) for s in data))
        return table



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
"""ModelArts service v1 action implementations"""
import logging

from osc_lib import exceptions
from osc_lib import utils
from osc_lib.cli import parseractions
from osc_lib.command import command

from otcextensions.common import cli_utils
from otcextensions.common import sdk_utils
from otcextensions.i18n import _

LOG = logging.getLogger(__name__)

INFER_TYPE_CHOICES = ["real-time", "batch"]

SORT_BY_CHOICES = ["publish_at", "service_name"]

STATUS_CHOICES = [
    "running",
    "deploying",
    "concerning",
    "failed",
    "stopped",
    "finished",
]

_formatters = {
    "additional_properties": cli_utils.YamlFormat,
    "config": cli_utils.YamlFormat,
    "operation_time": cli_utils.UnixTimestampFormatter,
    "publish_at": cli_utils.UnixTimestampFormatter,
    "transition_at": cli_utils.UnixTimestampFormatter,
    "update_time": cli_utils.UnixTimestampFormatter,
}


def _get_columns(item):
    column_map = {}
    hidden = ["location"]
    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden
    )


class ListServices(command.Lister):
    _description = _("Get properties of a service")
    columns = (
        "Service Id",
        "Service Name",
        "Infer Type",
        "Status",
    )

    def get_parser(self, prog_name):
        parser = super(ListServices, self).get_parser(prog_name)
        parser.add_argument(
            "--cluster-id",
            metavar="<cluster_id>",
            help=_("Dedicated resource pool ID."),
        )
        parser.add_argument(
            "--infer-type",
            metavar="{" + ",".join(INFER_TYPE_CHOICES) + "}",
            type=lambda s: s.lower(),
            choices=INFER_TYPE_CHOICES,
            help=_("Inference mode."),
        )
        parser.add_argument(
            "--limit",
            type=int,
            metavar="<limit>",
            help=_("Records Limit. Default value: 1000"),
        )
        parser.add_argument(
            "--model-id",
            metavar="<model_id>",
            help=_("Model Name."),
        )
        parser.add_argument(
            "--offset",
            metavar="<offset>",
            type=int,
            help=_("Start page of the paging list. Default value: 0"),
        )
        parser.add_argument(
            "--order",
            metavar="{asc, desc}",
            type=lambda s: s.lower(),
            choices=["asc", "desc"],
            help=_("Sorting order. Default value: desc"),
        )
        parser.add_argument(
            "--service-id",
            metavar="<service_id>",
            help=_("Service ID."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=_("Service Name."),
        )
        parser.add_argument(
            "--sort-by",
            metavar="{publish_at, service_name}",
            type=lambda s: s.lower(),
            choices=["publish_at", "service_name"],
            help=_("Sorting field. Default value: publish_at"),
        )
        parser.add_argument(
            "--status",
            metavar="{" + ",".join(STATUS_CHOICES) + "}",
            type=lambda s: s.lower(),
            choices=STATUS_CHOICES,
            help=_("Service status."),
        )
        parser.add_argument(
            "--workspace-id",
            metavar="<workspace_id>",
            help=_("Workspace ID."),
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1
        args_list = (
            "cluster_id",
            "infer_type",
            "limit",
            "model_id",
            "offset",
            "order",
            "service_id",
            "name",
            "sort_by",
            "status",
            "workspace_id",
        )
        query_params = {}
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val or str(val) == "0":
                query_params[arg] = val

        data = client.services(**query_params)

        return (
            self.columns,
            (utils.get_item_properties(s, self.columns) for s in data),
        )


class CreateService(command.ShowOne):
    _description = _("Create a ModelArts service")

    def get_parser(self, prog_name):
        parser = super(CreateService, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            metavar="<name>",
            help=_(
                "Service name. Enter 1 to 64 characters. Only letters, "
                "digits, hyphens (-), and underscores (_) are allowed."
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            required=False,
            help=_(
                "Service description, which contains a maximum of 100 "
                "characters. By default, this parameter is left blank."
            ),
        )
        parser.add_argument(
            "--infer-type",
            metavar="{" + ",".join(INFER_TYPE_CHOICES) + "}",
            type=lambda s: s.lower(),
            choices=INFER_TYPE_CHOICES,
            required=True,
            help=_("Inference mode."),
        )

        network_group = parser.add_argument_group("network arguments")
        parser._action_groups.insert(2, parser._action_groups.pop())
        network_group.add_argument(
            "--router-id",
            metavar="<router_id>",
            dest="vpc_id",
            help=_(
                "ID of the VPC to which a real-time service instance "
                "is deployed. By default, this parameter is left blank."
            ),
        )
        network_group.add_argument(
            "--network-id",
            metavar="<network_id>",
            dest="subnet_network_id",
            help=_(
                "ID of a subnet. By default, this parameter is left "
                "blank. This parameter is mandatory when vpc_id is "
                "configured. Enter the network ID displayed in the "
                "subnet details on the VPC management console "
            ),
        )
        network_group.add_argument(
            "--security-group-id",
            metavar="<security_group_id>",
            help=_(
                "A security group is a virtual firewall that provides "
                "secure network access control policies for service "
                "instances. "
            ),
        )
        parser.add_argument(
            "--cluster-id",
            metavar="<cluster_id>",
            required=False,
            help=_(
                "ID of a dedicated resource pool. If this parameter "
                "is set --router-id doesnt take effect."
            ),
        )
        parser.add_argument(
            "--schedule",
            metavar="type=<type>,time_unit=<time_unit>,duration=<duration>",
            required_keys=["type", "time_unit", "duration"],
            dest="schedule",
            action=parseractions.MultiKeyValueAction,
            help=_(
                "Service scheduling configuration, which can be "
                "configured only for real-time services."
            ),
        )
        parser.add_argument(
            "--workspace-id",
            metavar="<workspace_id>",
            help=_("Workspace ID."),
        )

        config_group = parser.add_argument_group("config arguments")
        parser._action_groups.insert(3, parser._action_groups.pop())
        config_group.add_argument(
            "--model-id",
            metavar="<model_id>",
            required=True,
            help=_("Model ID"),
        )
        config_group.add_argument(
            "--weight",
            metavar="<weight>",
            type=int,
            help=_(
                "Traffic weight allocated to a model. This parameter "
                "is mandatory only when infer_type is set to real-time. "
                "The sum of the weights must be 100."
            ),
        )
        config_group.add_argument(
            "--specification",
            metavar="<specification>",
            required=True,
            help=_(
                "Resource specifications. Select specifications based on "
                "service requirements."
            ),
        )
        config_group.add_argument(
            "--instance-count",
            metavar="<instance_count>",
            required=True,
            type=int,
            help=_(
                "Number of instances deployed in a model The value must "
                "be greater than 0."
            ),
        )
        config_group.add_argument(
            "--env",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="envs",
            help=_(
                "Environment variable key-value pair required "
                "for running a model.\n"
                "Example: --env VAR1=value1 --env VAR2=value2"
            ),
        )
        config_group.add_argument(
            "--custom-spec",
            metavar="cpu=<cpu>,memory=<memory>,gpu_p4=<gpu_p4>",
            required_keys=["cpu", "memory"],
            optional_keys=["gpu_p4"],
            dest="custom_spec",
            action=parseractions.MultiKeyValueAction,
            help=_(
                "Custom specifications. Set this parameter when "
                "you use a dedicated resource pool."
            ),
        )
        config_group.add_argument(
            "--src-type",
            metavar="<src_type>",
            help=_("Data source type."),
        )
        config_group.add_argument(
            "--src-path",
            metavar="<src_path>",
            help=_("OBS path of the input data of a batch job."),
        )
        config_group.add_argument(
            "--dest-path",
            metavar="<dest_path>",
            help=_("OBS path of the output data of a batch job."),
        )
        config_group.add_argument(
            "--req-uri",
            metavar="<dest_path>",
            help=_(
                "API URI from the model config.json file for inference. "
                "If a ModelArts built-in inference image is used, the "
                "value of this parameter is /"
            ),
        )
        config_group.add_argument(
            "--mapping-type",
            metavar="{file, csv}",
            type=lambda s: s.lower(),
            choices=["file", "csv"],
            help=_("Mapping type of the input data."),
        )
        config_group.add_argument(
            "--mapping-rule",
            metavar="<mapping_rule>",
            help=_(
                "Mapping between input parameters and CSV data. This "
                "parameter is mandatory only when mapping_type is set to csv."
            ),
        )
        parser.add_argument(
            "--wait",
            action="store_true",
            help=("Wait for service deployment."),
        )
        parser.add_argument(
            "--timeout",
            metavar="<timeout>",
            type=int,
            default=1200,
            help=_("Timeout for the wait in seconds (default 1200 seconds)."),
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1
        attrs = {
            "service_name": parsed_args.name,
            "infer_type": parsed_args.infer_type,
            "config": [
                {
                    "model_id": parsed_args.model_id,
                    "specification": parsed_args.specification,
                    "instance_count": parsed_args.instance_count,
                }
            ],
        }
        args_list = (
            "description",
            "workspace_id",
            "vpc_id",
            "subnet_network_id",
            "security_group_id",
            "cluster_id",
            "schedule",
        )
        for arg in args_list:
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val

        if parsed_args.envs:
            attrs["config"][0].update(envs=parsed_args.envs)

        if parsed_args.infer_type == "batch":
            for arg in ("src_path", "dest_path", "req_uri", "mapping_type"):
                val = getattr(parsed_args, arg)
                if val:
                    attrs["config"][0][arg] = val
                else:
                    raise exceptions.CommandError(
                        f"For batch infer_type --{arg.replace('_', '-')} "
                        "argument is required."
                    )
            if parsed_args.mapping_rule:
                attrs["config"][0]["mapping_rule"] = parsed_args.mapping_rule
            if parsed_args.src_type:
                attrs["config"][0]["src_type"] = parsed_args.src_type
            custom_spec = parsed_args.custom_spec
            if custom_spec:
                if len(custom_spec) > 1:
                    msg = "--custom-spec argument cannot be repeated"
                    raise exceptions.CommandError(msg)
                else:
                    custom_spec = custom_spec[0]
                    for key, val in custom_spec.items():
                        if key == "memory":
                            custom_spec[key] = int(val)
                        else:
                            custom_spec[key] = float(val)
                    attrs["config"][0].update(custom_spec=custom_spec)

        elif parsed_args.infer_type == "real-time":
            if parsed_args.weight:
                attrs["config"][0]["weight"] = parsed_args.weight
            else:
                raise exceptions.CommandError(
                    "For real-time infer_type --weight argument is required."
                )

        service = client.create_service(**attrs)
        if parsed_args.wait:
            client.wait_for_service(service.id, parsed_args.timeout)

        data = client.get_service(service.id)
        display_columns, columns = _get_columns(data)
        data = utils.get_item_properties(data, columns, formatters=_formatters)

        return display_columns, data


class UpdateService(command.ShowOne):
    _description = _("Update Service Configuration")

    def get_parser(self, prog_name):
        parser = super(UpdateService, self).get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            help=_("Service Name or ID"),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Service description (max 100 characters)."),
        )
        parser.add_argument(
            "--schedule",
            metavar="type=<type>,time_unit=<time_unit>,duration=<duration>",
            required_keys=["type", "time_unit", "duration"],
            dest="schedule",
            action=parseractions.MultiKeyValueAction,
            help=_(
                "Service scheduling configuration, which can be "
                "configured only for real-time services."
            ),
        )
        parser.add_argument(
            "--additional-property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="additional_properties",
            help=_(
                ".Additional service attribute, which facilitates service "
                "management\n"
                "Example: --additional-property VAR1=value1 "
                "--additional-property VAR2=value2"
            ),
        )

        config_group = parser.add_argument_group("config arguments")
        parser._action_groups.insert(2, parser._action_groups.pop())
        config_group.add_argument(
            "--model-id",
            metavar="<model_id>",
            help=_("Model ID"),
        )
        config_group.add_argument(
            "--weight",
            metavar="<weight>",
            type=int,
            help=_(
                "Traffic weight allocated to a model. This parameter "
                "is mandatory only when infer_type is set to real-time. "
                "The sum of the weights must be 100."
            ),
        )
        parser.add_argument(
            "--cluster-id",
            metavar="<cluster_id>",
            help=_(
                "ID of a dedicated resource pool. By default, this parameter "
                "is left blank, indicating that no dedicated resource "
                "pool is used."
            ),
        )
        config_group.add_argument(
            "--specification",
            metavar="<specification>",
            help=_(
                "Resource specifications. Select specifications based on "
                "service requirements."
            ),
        )
        config_group.add_argument(
            "--instance-count",
            metavar="<instance_count>",
            type=int,
            help=_(
                "Number of instances deployed in a model The value must "
                "be greater than 0."
            ),
        )
        config_group.add_argument(
            "--env",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="envs",
            help=_(
                "Environment variable key-value pair required "
                "for running a model.\n"
                "Example: --env VAR1=value1 --env VAR2=value2"
            ),
        )
        config_group.add_argument(
            "--custom-spec",
            metavar="cpu=<cpu>,memory=<memory>,gpu_p4=<gpu_p4>",
            required_keys=["cpu", "memory"],
            dest="custom_spec",
            action=parseractions.MultiKeyValueAction,
            help=_(
                "Custom specifications. Set this parameter when "
                "you use a dedicated resource pool."
            ),
        )
        config_group.add_argument(
            "--src-type",
            metavar="<src_type>",
            help=_("Data source type."),
        )
        config_group.add_argument(
            "--src-path",
            metavar="<src_path>",
            help=_("OBS path of the input data of a batch job."),
        )
        config_group.add_argument(
            "--dest-path",
            metavar="<dest_path>",
            help=_("OBS path of the output data of a batch job."),
        )
        config_group.add_argument(
            "--req-uri",
            metavar="<dest_path>",
            help=_(
                "API URI from the model config.json file for inference. "
                "If a ModelArts built-in inference image is used, the "
                "value of this parameter is /"
            ),
        )
        config_group.add_argument(
            "--mapping-type",
            metavar="{file, csv}",
            type=lambda s: s.lower(),
            choices=["file", "csv"],
            help=_("Mapping type of the input data."),
        )
        config_group.add_argument(
            "--mapping-rule",
            metavar="<mapping_rule>",
            help=_(
                "Mapping between input parameters and CSV data. This "
                "parameter is mandatory only when mapping_type is set to csv."
            ),
        )

        parser.add_argument(
            "--wait",
            action="store_true",
            help=("Wait for service deployment."),
        )
        parser.add_argument(
            "--timeout",
            metavar="<timeout>",
            type=int,
            default=1200,
            help=_("Timeout for the wait in seconds (default 1200 seconds)."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1

        config = {}
        for arg in ("model_id", "specification", "instance_count", "envs"):
            val = getattr(parsed_args, arg)
            if val:
                config[arg] = val

        service = client.find_service(
            parsed_args.service, ignore_missing=False
        )
        if service.infer_type == "real-time":
            for arg in ("weight", "cluster_id"):
                val = getattr(parsed_args, arg)
                if val:
                    config[arg] = val

            custom_spec = parsed_args.custom_spec
            if custom_spec:
                if len(custom_spec) > 1:
                    msg = "--custom-spec argument cannot be repeated"
                    raise exceptions.CommandError(msg)
                else:
                    custom_spec = custom_spec[0]
                    for key, val in custom_spec.items():
                        if key == "memory":
                            custom_spec[key] = int(val)
                        else:
                            custom_spec[key] = float(val)
                    config.update(custom_spec=custom_spec)

        elif service.infer_type == "batch":
            for arg in (
                "src_path",
                "dest_path",
                "req_uri",
                "mapping_type",
                "mapping_value",
            ):
                val = getattr(parsed_args, arg)
                if val:
                    config[arg] = val

        attrs = {}
        for arg in ("description", "schedule", "additional_properties"):
            val = getattr(parsed_args, arg)
            if val:
                attrs[arg] = val

        attrs.update(config=[config])

        client.update_service(service.id, **attrs)
        if parsed_args.wait:
            client.wait_for_service(service.id, parsed_args.timeout)

        data = client.get_service(service.id)
        display_columns, columns = _get_columns(data)
        data = utils.get_item_properties(data, columns, formatters=_formatters)

        return display_columns, data


class ShowService(command.ShowOne):
    _description = _("Show details of a Modelarts serivce")

    def get_parser(self, prog_name):
        parser = super(ShowService, self).get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            help=_("Service Name or ID"),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1

        service = client.find_service(
            parsed_args.service, ignore_missing=False
        )
        data = client.get_service(service.id)

        display_columns, columns = _get_columns(data)

        data = utils.get_item_properties(data, columns, formatters=_formatters)
        return (display_columns, data)


class StartService(command.Command):
    _description = _("Start a Service.")

    def get_parser(self, prog_name):
        parser = super(StartService, self).get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            help=_("Service name or ID."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1

        obj = client.find_service(parsed_args.service, ignore_missing=False)
        client.start_service(obj)


class StopService(command.Command):
    _description = _("Stop a Service.")

    def get_parser(self, prog_name):
        parser = super(StopService, self).get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            help=_("Service name or ID."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1

        obj = client.find_service(parsed_args.service, ignore_missing=False)
        client.stop_service(obj)


class DeleteService(command.Command):
    _description = _("Delete ModelArts Service(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteService, self).get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            nargs="+",
            help=_("ID or Name of the ModelArts service(s) to be deleted."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.modelartsv1
        result = 0
        for name_or_id in parsed_args.service:
            try:
                service = client.find_service(name_or_id, ignore_missing=False)
                client.delete_service(service.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete service(s) with "
                        "ID or Name '%(service)s': %(e)s"
                    ),
                    {"service": name_or_id, "e": e},
                )
        if result > 0:
            total = len(parsed_args.service)
            msg = _(
                "%(result)s of %(total)s Service(s) failed " "to delete."
            ) % {
                "result": result,
                "total": total,
            }
            raise exceptions.CommandError(msg)
